"""
Utility for validating a given classifier implementation's model against
some labeled testing data, outputting PR and ROC curve plots with
area-under-curve score values.

This utility can optionally be used train a supervised classifier model if
the given classifier model configuration does not exist and a second CSV
file listing labeled training data is provided. Training will be attempted
if ``train`` is set to true. If training is performed, we exit after
training completes. A ``SupervisedClassifier`` sub-classing implementation
must be configured

We expect the test and train CSV files in the column format:

    ...
    <UUID>,<label>
    ...

The UUID is of the descriptor to which the label applies. The label may be
any arbitrary string value, but all labels must be consistent in
application.

Some metrics presented assume the highest confidence class as the single
predicted class for an element:

    - confusion matrix

The output UUID confusion matrix is a JSON dictionary where the top-level
keys are the true labels, and the inner dictionary is the mapping of
predicted labels to the UUIDs of the classifications/descriptors that
yielded the prediction. Again, this is based on the maximum probability
label for a classification result (T=0.5).

See Scikit-Learn PR and ROC curve explanations and examples:
    - http://scikit-learn.org/stable/auto_examples/model_selection/plot_precision_recall.html
    - http://scikit-learn.org/stable/auto_examples/model_selection/plot_roc.html

"""
from __future__ import division, print_function

import collections
import csv
import json
import logging

import matplotlib.pyplot as plt
import numpy
import scipy.stats
import six
from six.moves import range
import sklearn.metrics

from smqtk.algorithms import (
    get_classifier_impls,
    SupervisedClassifier,
)
from smqtk.representation import (
    ClassificationElementFactory,
    get_descriptor_index_impls,
)
from smqtk.utils import (
    bin_utils,
    parallel,
    plugin,
)


__author__ = "paul.tunison@kitware.com"


def default_config():
    return {
        'plugins': {
            'classifier':
                plugin.make_config(get_classifier_impls()),
            'classification_factory':
                ClassificationElementFactory.get_default_config(),
            'descriptor_index':
                plugin.make_config(get_descriptor_index_impls())
        },
        'utility': {
            'train': False,
            'csv_filepath': 'CHAMGEME :: PATH :: a csv file',
            'output_plot_pr': None,
            'output_plot_roc': None,
            'output_plot_confusion_matrix': None,
            'output_uuid_confusion_matrix': None,
            'curve_confidence_interval': False,
            'curve_confidence_interval_alpha': 0.4,
        },
        "parallelism": {
            "descriptor_fetch_cores": 4,
            "classification_cores": None,
        },
    }


def cli_parser():
    return bin_utils.basic_cli_parser(__doc__)


def main():
    args = cli_parser().parse_args()
    config = bin_utils.utility_main_helper(default_config, args)
    log = logging.getLogger(__name__)

    #
    # Initialize stuff from configuration
    #
    #: :type: smqtk.algorithms.Classifier
    classifier = plugin.from_plugin_config(
        config['plugins']['classifier'],
        get_classifier_impls()
    )
    #: :type: ClassificationElementFactory
    classification_factory = ClassificationElementFactory.from_config(
        config['plugins']['classification_factory']
    )
    #: :type: smqtk.representation.DescriptorIndex
    descriptor_index = plugin.from_plugin_config(
        config['plugins']['descriptor_index'],
        get_descriptor_index_impls()
    )

    uuid2label_filepath = config['utility']['csv_filepath']
    do_train = config['utility']['train']
    output_uuid_cm = config['utility']['output_uuid_confusion_matrix']
    plot_filepath_pr = config['utility']['output_plot_pr']
    plot_filepath_roc = config['utility']['output_plot_roc']
    plot_filepath_cm = config['utility']['output_plot_confusion_matrix']
    plot_ci = config['utility']['curve_confidence_interval']
    plot_ci_alpha = config['utility']['curve_confidence_interval_alpha']

    #
    # Construct mapping of label to the DescriptorElement instances for that
    # described by that label.
    #
    log.info("Loading descriptors by UUID")

    def iter_uuid_label():
        """ Iterate through UUIDs in specified file """
        with open(uuid2label_filepath) as uuid2label_file:
            reader = csv.reader(uuid2label_file)
            for r in reader:
                # TODO: This will need to be updated to handle multiple labels
                #       per descriptor.
                yield r[0], r[1]

    def get_descr(r):
        """ Fetch descriptors from configured index """
        uuid, truth_label = r
        return truth_label, descriptor_index.get_descriptor(uuid)

    tlabel_element_iter = parallel.parallel_map(
        get_descr, iter_uuid_label(),
        name="cmv_get_descriptors",
        use_multiprocessing=True,
        cores=config['parallelism']['descriptor_fetch_cores'],
    )

    # Map of truth labels to descriptors of labeled data
    #: :type: dict[str, list[smqtk.representation.DescriptorElement]]
    tlabel2descriptors = {}
    for tlabel, d in tlabel_element_iter:
        tlabel2descriptors.setdefault(tlabel, []).append(d)

    # Train classifier if the one given has a ``train`` method and training
    # was turned enabled.
    if do_train:
        if isinstance(classifier, SupervisedClassifier):
            log.info("Training classifier model")
            classifier.train(tlabel2descriptors)
            exit(0)
        else:
            ValueError("Configured classifier is not a SupervisedClassifier "
                       "type and does not support training.")

    #
    # Apply classifier to descriptors for predictions
    #

    # Truth label to predicted classification results
    #: :type: dict[str, set[smqtk.representation.ClassificationElement]]
    tlabel2classifications = {}
    for tlabel, descriptors in six.iteritems(tlabel2descriptors):
        tlabel2classifications[tlabel] = \
            set(classifier.classify_async(
                descriptors, classification_factory,
                use_multiprocessing=True,
                procs=config['parallelism']['classification_cores'],
                ri=1.0,
            ).values())
    log.info("Truth label counts:")
    for l in sorted(tlabel2classifications):
        log.info("  %s :: %d", l, len(tlabel2classifications[l]))

    #
    # Confusion Matrix
    #
    conf_mat, labels = gen_confusion_matrix(tlabel2classifications)
    log.info("Confusion_matrix")
    log_cm(log.info, conf_mat, labels)
    if plot_filepath_cm:
        plot_cm(conf_mat, labels, plot_filepath_cm)

    # Confusion Matrix of descriptor UUIDs to output json
    if output_uuid_cm:
        # Top dictionary keys are true labels, inner dictionary keys are UUID
        # predicted labels.
        log.info("Computing UUID Confusion Matrix")
        #: :type: dict[str, dict[collections.Hashable, set]]
        uuid_cm = {}
        for tlabel in tlabel2classifications:
            uuid_cm[tlabel] = collections.defaultdict(set)
            for c in tlabel2classifications[tlabel]:
                uuid_cm[tlabel][c.max_label()].add(c.uuid)
            # convert sets to lists
            for plabel in uuid_cm[tlabel]:
                uuid_cm[tlabel][plabel] = list(uuid_cm[tlabel][plabel])
        with open(output_uuid_cm, 'w') as f:
            log.info("Saving UUID Confusion Matrix: %s", output_uuid_cm)
            json.dump(uuid_cm, f, indent=2, separators=(',', ': '))

    #
    # Create PR/ROC curves via scikit learn tools
    #
    if plot_filepath_pr:
        log.info("Making PR curve")
        make_pr_curves(tlabel2classifications, plot_filepath_pr,
                       plot_ci, plot_ci_alpha)
    if plot_filepath_roc:
        log.info("Making ROC curve")
        make_roc_curves(tlabel2classifications, plot_filepath_roc,
                        plot_ci, plot_ci_alpha)


def gen_confusion_matrix(tlabel2classifications):
    """
    Generate numpy confusion matrix based on classification highest confidence
    score.

    :param tlabel2classifications: Mapping of true label for mapped set of
        classifications.
    :type tlabel2classifications:
        dict[str, set[smqtk.representation.ClassificationElement]]

    :return: Numpy confusion matrix and label vectors for rows and columns
    :rtype: numpy.ndarray[int], list[str]

    """
    # List of true and predicted classes for classifications
    true_classes = []
    pred_classes = []

    for true_label in tlabel2classifications:
        for c in tlabel2classifications[true_label]:
            true_classes.append(true_label)
            pred_classes.append(c.max_label())

    labels = sorted(set(true_classes).union(pred_classes))
    confusion_mat = sklearn.metrics.confusion_matrix(true_classes,
                                                     pred_classes,
                                                     labels)

    return confusion_mat, labels


def log_cm(p_func, conf_mat, labels):
    print_mat = numpy.zeros((conf_mat.shape[0] + 2, conf_mat.shape[1] + 1),
                            dtype=object)
    print_mat[0, 0] = "Predicted"
    print_mat[1, 0] = "Actual"
    print_mat[2:, 1:] = conf_mat.astype(str)
    print_mat[0, 1:] = labels
    print_mat[1, 1:] = ''
    print_mat[2:, 0] = labels

    # get col max widths
    col_max_lens = []
    for x in range(print_mat.shape[1]):
        col_max_lens.append(max(list(map(len, print_mat[:, x].flatten().tolist()))))

    # Construct printed rows based on column max width
    p_func("Confusion Matrix (Counts)")
    for r in print_mat:
        segs = []
        for i, w in enumerate(r):
            segs.append(' ' * (col_max_lens[i] - len(w)) + w)
        p_func(' '.join(segs))


def plot_cm(conf_mat, labels, output_path):
    """
    :param conf_mat: Confusion matrix with items counts
    :type conf_mat: numpy.ndarray

    :param labels: Symmetric row/column labels
    :type labels: list[str]

    :param output_path: Path to save generated figure.
    :type output_path: str

    :return:
    :rtype:

    """
    log = logging.getLogger(__name__)

    #: :type: numpy.ndarray
    cm = conf_mat.copy()
    log.debug("raw conf mat:\n%s", cm)

    # each row represents a true class
    cm_f = cm / cm.sum(1).astype(float)[:, numpy.newaxis]
    log.debug("normalized conf mat:\n%s", cm_f)

    fig = plt.figure()
    #: :type: matplotlib.axes._axes.Axes
    ax = fig.add_subplot(111)
    cax = ax.matshow(cm_f, vmin=0.0, vmax=1.0)
    fig.colorbar(cax)

    # Annotate cells with count values
    for y in range(cm.shape[0]):
        for x in range(cm.shape[1]):
            ax.annotate(s=str(cm[y, x]), xy=(x, y), xycoords='data')

    ax.set_xticklabels([''] + labels)
    ax.set_yticklabels([''] + labels)
    ax.set_title('Confusion Matrix - Percent Makeup')
    ax.set_xlabel('Predicted Class')
    ax.set_ylabel('True Class')
    fig.savefig(output_path)


def format_plt(title, x_label, y_label):
    plt.xlim([0., 1.])
    plt.ylim([0., 1.05])
    plt.xlabel(x_label)
    plt.ylabel(y_label)
    plt.title(title)
    plt.legend(loc='best', fancybox=True, framealpha=0.5)


def select_color_marker(i):
    """ Return index-based marker/color format for plotting """
    colors = ['b', 'g', 'r', 'c', 'y', 'k']
    style = ['-', '--', '-.', ':']
    ci = i % len(colors)
    si = (i // len(colors)) % len(style)
    return '%s%s' % (colors[ci], style[si])


def make_curve(log, skl_curve_func, title, xlabel, ylabel, output_filepath,
               label2classifications, plot_ci, plot_ci_alpha,
               plot_ci_make_poly):
    """
    :param log: Logger to use.
    :type log: logging.Logger

    :param skl_curve_func: scikit-learn curve generation function. This should
        be wrapped to return (x, y) value arrays of the curve plot.
    :type skl_curve_func:
        (numpy.ndarray[float], numpy.ndarray[float]) ->
            (numpy.ndarray[float], numpy.ndarray[float], numpy.ndarray[float])

    :param title: Title of the plot.
    :param xlabel: X-axis label for the plot.
    :param ylabel: Y-axis label for the plot.
    :param output_filepath: Path to write the generated plot image to.

    :param label2classifications: Mapping of label to the classification
        elements that should be that label.
    :type label2classifications:
        dict[str, set[smqtk.representation.ClassificationElement]]

    :param plot_ci: Flag for whether to draw the confidence interval or not.
    :type plot_ci: bool

    :param plot_ci_alpha: Alpha value to use for coloring the confidence
        interval area in the range [0, 1].
    :type plot_ci_alpha: float

    :param plot_ci_make_poly: Function that takes x, y, and their upper and
        lower confidence interval estimations, and returns a plt.Polygon
        correctly for the curve we're trying to draw.
    :type plot_ci_make_poly: (x, x_l, x_u, y, y_l, y_u, **poly_args) ->
        plt.Polygon

    """
    # Create curves for each label and then for overall.

    all_classifications = set()
    for s in label2classifications.values():
        all_classifications.update(s)

    plt.clf()
    plt.figure(figsize=(15, 12), dpi=72)
    for i, l in enumerate(sorted(label2classifications)):
        # record binary truth relation with respect to label `l`
        l_truth = []
        l_proba = []
        # Get classification probability and truth label for all tested elements
        for c in all_classifications:
            l_truth.append(int(c in label2classifications[l]))
            l_proba.append(c[l])
        assert len(l_truth) == len(l_proba) == len(all_classifications), \
            "Somehow didn't wind up with truth/proba values for all " \
            "classification elements"
        x, y, t = skl_curve_func(numpy.array(l_truth), numpy.array(l_proba))
        auc = sklearn.metrics.auc(x, y)
        m = select_color_marker(i)
        plt.plot(x, y, m, label="%s (auc=%f)" % (l, auc))

        if plot_ci:
            # Confidence interval calculation using Wilson's score interval
            x_u, x_l = curve_wilson_ci(x, len(l_proba))
            y_u, y_l = curve_wilson_ci(y, len(l_proba))

            # For each xy point, combine CI upper/lower bounds based on curve
            # being drawn. I.e. zip(x_l, y_l) for PR curve, but zip(x_u, y_l)
            # for ROC curve.
            # TODO: Generate concave hull (alpha shape)
            #       This would be curve direction independent.
            ci_poly = plot_ci_make_poly(x, x_l, x_u, y, y_l, y_u,
                                        facecolor=m[0], edgecolor=m[0],
                                        alpha=plot_ci_alpha)
            plt.gca().add_patch(ci_poly)

    format_plt(title + ' - Class vs. Rest', xlabel, ylabel)
    log.info("Saving: %s", output_filepath)
    plt.savefig(output_filepath)
    plt.clf()


def make_pr_curves(label2classifications, output_filepath, plot_ci,
                   plot_ci_alpha):
    def skl_pr_curve(truth, proba):
        p, r, t = sklearn.metrics.precision_recall_curve(truth, proba,
                                                         pos_label=1)
        return r, p, t

    def make_ci_poly(x, x_l, x_u, y, y_l, y_u, **poly_kwds):
        x_l = numpy.min([x, x_l], 0)
        x_u = numpy.max([x, x_u], 0)
        y_l = numpy.min([y, y_l], 0)
        y_u = numpy.max([y, y_u], 0)
        # Add points to flush ends with plot border
        poly_points = (
            [(x[0], y_l[0])] + zip(x_l, y_l) + [(x[-1], y_l[-1])] +
            [(x[-1], y_u[-1])] + list(reversed(zip(x_u, y_u))) +
            [(x[0], y_u[0])]
        )
        return plt.Polygon(poly_points, **poly_kwds)

    log = logging.getLogger(__name__)
    make_curve(log, skl_pr_curve, "PR", "Recall", "Precision", output_filepath,
               label2classifications, plot_ci, plot_ci_alpha, make_ci_poly)


def make_roc_curves(label2classifications, output_filepath, plot_ci,
                    plot_ci_alpha):
    def skl_roc_curve(truth, proba):
        fpr, tpr, t = sklearn.metrics.roc_curve(truth, proba, pos_label=1)
        return fpr, tpr, t

    def make_ci_poly(x, x_l, x_u, y, y_l, y_u, **poly_kwds):
        x_l = numpy.min([x, x_l], 0)
        x_u = numpy.max([x, x_u], 0)
        y_l = numpy.min([y, y_l], 0)
        y_u = numpy.max([y, y_u], 0)
        poly_points = (
            [(x[0], y_u[0])] + zip(x_l, y_u) + [(x[-1], y_u[-1])] +
            [(x[-1], y_l[-1])] + list(reversed(zip(x_u, y_l))) +
            [(x[0], y_l[0])]
        )
        return plt.Polygon(poly_points, **poly_kwds)

    log = logging.getLogger(__name__)
    make_curve(log, skl_roc_curve, "ROC", "False Positive Rate",
               "True Positive Rate", output_filepath, label2classifications,
               plot_ci, plot_ci_alpha, make_ci_poly)


def curve_wilson_ci(p, n, confidence=0.95):
    """
    Generate upper and lower bounds confidence interval, using the Wilson score
    interval, for PR and ROC curves.

    :param p: Array of points for an axis.
    :type p: numpy.ndarray[float]

    :param n: number of source predictions that led to the ``p`` array.
    :type n: int

    :param confidence: (0, 1) confidence value for interval generation. Default
        of 0.95, or 95% confidence bounds.
    :type confidence: float

    :return: Upper and lower bounds arrays.
    :rtype: (numpy.ndarray[float], numpy.ndarray[float])

    """
    quantile = 1.-(0.5*(1.-confidence))
    z = scipy.stats.norm.ppf(quantile)  # ~1.96 when confidence==0.95
    n = float(n)
    u = (1/(1+(z*z/n))) * \
        (p + (z*z)/(2*n) + z*numpy.sqrt(p * (1 - p) / n + (z*z)/(4*n*n)))
    L = (1/(1+(z*z/n))) * \
        (p + (z*z)/(2*n) - z*numpy.sqrt(p * (1 - p) / n + (z*z)/(4*n*n)))
    return u, L


if __name__ == '__main__':
    main()



import atexit
import threading
import time

from smqtk.utils import SmqtkObject


class IqrController (SmqtkObject):
    """
    Main controlling object for one or more IQR Sessions.

    In order to interface with a web server, methods defined here are
    thread-safe.

    This class may be used with the ``with`` statement. This will enable the
    instance's primary lock, preventing any other action from being performed
    on the instance while inside the with statement. The lock is reentrant, so
    nested with-statements will not dead-lock.

    """

    def __init__(self, expire_enabled=False, expire_check=30,
                 expire_callback=None):
        """
        Initialize the controller.

        Session timeout, when enabled, is set on a session-by-session basis,
        i.e. different session may have different time out values if desired.
        When expiration is not enabled, session timeout values are ignored.

        :param expire_enabled: Enable/Disable session expiry. If enabled, a
            thread is started for monitoring and removal.
        :type expire_enabled:

        :param expire_check: Interval, in seconds, that we check for session
            expiration.
        :type expire_check: float

        :param expire_callback: Optional callable that should take one
            positional parameter, the session that is expiring, and is called
            when the session expires and just before it is removed from this
            controller.

            The provided function, when called, will be within this controller's
            lock.

            If expiration is NOT enabled, or if a session is not given a
            timeout, this callback function is not used.
        :type expire_callback: (collections.Hashable) -> None

        """
        # Map of uuid to the search state
        #: :type: dict[collections.Hashable, IqrSession]
        self._iqr_sessions = {}
        # Map of sessions with timeout's enabled and the time out value in
        # seconds
        #: :type
        self._iqr_session_timeout = {}
        # Map of the UNIX time a session was last accessed
        self._iqr_session_last_access = {}

        # RLock for iqr_session[*] maps.
        self._map_rlock = threading.RLock()

        self._expire_enabled = expire_enabled
        self._expire_interval = expire_check
        self._expire_thread_stop_event = threading.Event()
        # prevents calling _handle_session_expiration when not enabled
        self._expire_thread_stop_event.set()
        self._expire_thread = None
        self._expire_callback = expire_callback

        # If enabled, start expiration monitor thread
        if self._expire_enabled:
            atexit.register(self.stop_expiration_monitor)
            self.start_expiration_monitor()

    def __enter__(self):
        self._map_rlock.acquire()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._map_rlock.release()

    def _handle_session_expiration(self):
        """
        Run on a separate thread periodic checks and removals of sessions that
        have expired.
        """
        while not self._expire_thread_stop_event.wait(self._expire_interval):
            # scan timeout-set sessions, removing those that have timed out
            now = time.time()
            with self._map_rlock:
                self._log.debug("Checking session expiration timeouts")
                for sid in self._iqr_session_timeout.keys():
                    to = self._iqr_session_timeout[sid]
                    la = self._iqr_session_last_access[sid]
                    t = now - la
                    if t > to:
                        self._log.debug("-> Expiring session '%s' "
                                        "(last-access: %s, timeout: %s, "
                                        "now: %s)", sid, la, to, now)
                        if hasattr(self._expire_callback, '__call__'):
                            self._log.debug("   - Executing callback")
                            self._expire_callback(self._iqr_sessions[sid])
                        self.remove_session(sid)

        self._log.debug("End of expiration handle function")

    def start_expiration_monitor(self):
        """
        Initialize unique thread to check for session expiration if the feature
        is enabled. This does nothing if the feature is not enabled.

        We stop the previous thread if one was started.

        """
        with self._map_rlock:
            self.stop_expiration_monitor()

            if self._expire_enabled:
                self._log.debug("Starting session expiration monitor thread")
                self._expire_thread = threading.Thread(
                    target=self._handle_session_expiration
                )
                self._expire_thread.daemon = True
                self._expire_thread_stop_event.clear()
                self._expire_thread.start()

    def stop_expiration_monitor(self):
        """
        Stop the session expiration monitoring thread if one has been started.
        Otherwise this method does nothing.
        """
        with self._map_rlock:
            if self._expire_thread:
                self._log.debug("Stopping session expiration monitor thread")
                self._expire_thread_stop_event.set()
                self._expire_thread.join()
                self._expire_thread = None
                self._log.debug("Stopping session expiration monitor thread "
                                "-- Done")

    def session_uuids(self):
        """
        Return a tuple of all currently registered IqrSessions.

        This does NOT update session last access in regards to session
        expiration.

        :return: a tuple of all currently registered IqrSessions.
        :rtype: tuple of collections.Hashable

        """
        with self._map_rlock:
            return tuple(self._iqr_sessions)

    def has_session_uuid(self, session_uuid):
        """ Check if this controller contains a session referenced by the given
        ID.

        Performance using this function is faster compared to getting all UUIDs
        and performing a linear search (because hash tables).

        This does NOT update session last access in regards to session
        expiration.

        :param session_uuid: Possible UUID of a session
        :type session_uuid: collections.Hashable

        :return: True of the given UUID references a session in this controller
            and false if not.
        :rtype: bool

        """
        with self._map_rlock:
            return session_uuid in self._iqr_sessions

    def add_session(self, iqr_session, timeout=0):
        """ Initialize a new IQR Session, returning the uuid of that session

        This controller indexes the given session by its UUID.

        :param iqr_session: The IqrSession instance to add
        :type iqr_session: smqtk.iqr.iqr_session.IqrSession

        :param timeout: The optional timeout, in seconds.
        :type timeout: float

        :return: UUID of new IQR Session
        :rtype: collections.Hashable

        """
        timeout = float(timeout)
        with self._map_rlock:
            sid = iqr_session.uuid
            if sid in self._iqr_sessions:
                raise RuntimeError("Cannot use given session as its UUID "
                                   "already exists in the controller session "
                                   "map: %s" % sid)

            self._iqr_sessions[sid] = iqr_session
            if timeout > 0:
                self._iqr_session_timeout[sid] = timeout
                self._iqr_session_last_access[sid] = time.time()
            return sid

    def get_session(self, session_uuid):
        """
        Return the session instance for the given UUID

        :raises KeyError: The given UUID doesn't exist in this controller.

        :param session_uuid: UUID if the session to get
        :type session_uuid: collections.Hashable

        :return: IqrSession instance for the given UUID
        :rtype: smqtk.iqr.iqr_session.IqrSession

        """
        with self._map_rlock:
            if session_uuid in self._iqr_session_timeout:
                self._iqr_session_last_access[session_uuid] = time.time()
            return self._iqr_sessions[session_uuid]

    def remove_session(self, session_uuid):
        """
        Remove an IQR Session by session UUID.

        :raises KeyError: The given UUID doesn't exist in this controller.

        :param session_uuid: Session UUID
        :type session_uuid: collections.Hashable

        """
        with self._map_rlock:
            del self._iqr_sessions[session_uuid]
            if session_uuid in self._iqr_session_timeout:
                del self._iqr_session_timeout[session_uuid]
                del self._iqr_session_last_access[session_uuid]



from __future__ import division, print_function
import unittest

import mock

from smqtk.algorithms.nn_index.lsh.functors import LshFunctor


class DummyLshFunctor (LshFunctor):

    @classmethod
    def is_usable(cls):
        return True

    def get_config(self):
        pass

    def get_hash(self, descriptor):
        pass


class testLshFunctorAbstract (unittest.TestCase):

    def test_call(self):
        # calling an instance should get us to the get_hash method.
        f = DummyLshFunctor()
        f.get_hash = mock.MagicMock()

        expected_descriptor = 'pretend descriptor element'
        f(expected_descriptor)
        f.get_hash.assert_called_once_with(expected_descriptor)



import unittest

import mock
import six
from six.moves import cPickle as pickle

from smqtk.exceptions import ReadOnlyError
from smqtk.representation.data_element.memory_element import DataMemoryElement
from smqtk.representation.key_value.memory import MemoryKeyValueStore


class TestMemoryKeyValueStore (unittest.TestCase):

    def test_is_usable(self):
        # Should always be usable
        self.assertTrue(MemoryKeyValueStore.is_usable())

    def test_get_default(self):
        # Check default config
        default_config = MemoryKeyValueStore.get_default_config()
        self.assertIsInstance(default_config, dict)
        # - Should just contain cache element property, which is a nested
        #   plugin config with no default type.
        self.assertIn('cache_element', default_config)
        self.assertIn('type', default_config['cache_element'])
        self.assertIsNone(default_config['cache_element']['type'])

    def test_from_config_empty_config(self):
        # should resort to default parameters
        s = MemoryKeyValueStore.from_config({})
        self.assertIsNone(s._cache_element)
        self.assertEqual(s._table, {})

    def test_from_config_none_value(self):
        # When cache_element value is None
        expected_config = {'cache_element': None}
        s = MemoryKeyValueStore.from_config(expected_config)
        self.assertIsNone(s._cache_element)
        self.assertEqual(s._table, {})

    def test_from_config_none_type(self):
        # When config map given, but plugin type set to null/None
        config = {'cache_element': {
            'some_type': {'param': None},
            'type': None,
        }}
        s = MemoryKeyValueStore.from_config(config)
        self.assertIsNone(s._cache_element)
        self.assertEqual(s._table, {})

    def test_from_config_with_cache_element(self):
        # Pickled dictionary with a known entry
        expected_table = {'some_key': 'some_value'}
        empty_dict_pickle = six.b("(dp1\nS'some_key'\np2\nS'some_value'\np3\ns.")

        # Test construction with memory data element.
        config = {'cache_element': {
            'DataMemoryElement': {
                'bytes': empty_dict_pickle,
            },
            'type': 'DataMemoryElement'
        }}
        s = MemoryKeyValueStore.from_config(config)
        self.assertIsInstance(s._cache_element, DataMemoryElement)
        self.assertEqual(s._table, expected_table)

    def test_new_no_cache(self):
        """ Test construction with no cache element. """
        s = MemoryKeyValueStore()
        self.assertIsNone(s._cache_element)
        self.assertEqual(s._table, {})

    def test_new_empty_cache(self):
        """
        Test construction with a cache element set with no bytes (empty).
        """
        c = DataMemoryElement()
        s = MemoryKeyValueStore(c)
        self.assertEqual(s._cache_element, c)
        self.assertEqual(s._table, {})

    def test_new_cached_table(self):
        """
        Test construction with a cached table.
        """
        expected_table = {
            'a': 'b',
            'c': 1,
            'asdfghsdfg': None,
            'r3adf3a#+': [4, 5, 6, '7'],
        }
        expected_table_pickle = pickle.dumps(expected_table, 2)

        c = DataMemoryElement(expected_table_pickle)
        s = MemoryKeyValueStore(c)
        self.assertEqual(s._cache_element, c)
        self.assertEqual(s._table, expected_table)

    def test_repr_no_cache(self):
        """
        Test representational string when no cache is set.
        """
        expected_repr = '<MemoryKeyValueStore cache_element: None>'
        s = MemoryKeyValueStore()
        actual_repr = repr(s)
        self.assertEqual(actual_repr, expected_repr)

    def test_repr_simple_cache(self):
        """
        Test representational string when a cache element is set.
        """
        c = DataMemoryElement()
        s = MemoryKeyValueStore(c)
        expected_repr = "<MemoryKeyValueStore cache_element: " \
                        "DataMemoryElement{len(bytes): 0, content_type: " \
                        "None, readonly: False}>"
        self.assertEqual(repr(s), expected_repr)

    def test_count(self):
        """
        Test that count returns appropriately based on table state.
        """
        s = MemoryKeyValueStore()
        assert s.count() == 0
        s._table = {
            0: 0,
            1: 1,
            'a': True,
            None: False
        }
        assert s.count() == 4

    def test_get_config_no_cache_elem(self):
        """
        Test that configuration returned reflects no cache element being set.
        """
        s = MemoryKeyValueStore()
        s._cache_element = None
        # We expect an default DataElement config (no impl type defined)
        c = s.get_config()
        self.assertIn('cache_element', c)
        self.assertIsNone(c['cache_element']['type'])

    def test_get_config_mem_cache_elem(self):
        """
        Test that configuration returned reflects the cache element that is
        set.
        """
        s = MemoryKeyValueStore()
        s._cache_element = DataMemoryElement(six.b('someBytes'), 'text/plain', False)
        expected_config = {'cache_element': {
            "DataMemoryElement": {
                'bytes': six.b('someBytes'),
                'content_type': 'text/plain',
                'readonly': False,
            },
            'type': 'DataMemoryElement'
        }}
        self.assertEqual(s.get_config(), expected_config)

    def test_keys_empty(self):
        """
        Test that no keys are present in a freshly constructed instance.
        """
        s = MemoryKeyValueStore()
        self.assertEqual(list(s.keys()), [])

    def test_keys_with_table(self):
        """
        Test that keys returned reflect the table state.
        """
        s = MemoryKeyValueStore()
        s._table = {
            'a': 'b',
            'c': 1,
            'asdfghsdfg': None,
            'r3adf3a#+': [4, 5, 6, '7'],
        }
        self.assertSetEqual(
            set(s.keys()),
            {'a', 'c', 'asdfghsdfg', 'r3adf3a#+'}
        )

    def test_read_only_no_cache(self):
        """ Test that read_only is false when there is no cache. """
        s = MemoryKeyValueStore()
        self.assertIsNone(s._cache_element)
        self.assertFalse(s.is_read_only())

    def test_read_only_with_writable_cache(self):
        """ Test that read-only status reflects a non-read-only cache """
        s = MemoryKeyValueStore()
        s._cache_element = DataMemoryElement(readonly=False)
        self.assertFalse(s.is_read_only())

    def test_read_only_with_read_only_cache(self):
        """ Test that read-only status reflects a read-only cache """
        s = MemoryKeyValueStore()
        s._cache_element = DataMemoryElement(readonly=True)
        self.assertTrue(s.is_read_only())

    def test_has_invalid_key(self):
        """ Test that has-key is false for a key not in the table. """
        s = MemoryKeyValueStore()
        self.assertEqual(s._table, {})
        self.assertFalse(s.has('some key'))

    def test_has_key(self):
        """ Test that has-key returns true for entered keys. """
        s = MemoryKeyValueStore()
        s._table = {
            'a': 0,
            'b': 1,
            0: 2,
        }
        self.assertTrue(s.has('a'))
        self.assertTrue(s.has('b'))
        self.assertTrue(s.has(0))
        self.assertFalse(s.has('c'))

    def test_add_invalid_key(self):
        """ Test that we cannot add non-hashable keys. """
        s = MemoryKeyValueStore()

        self.assertRaises(TypeError, s.add, [1, 2, 3], 0)
        self.assertEqual(s._table, {})

        self.assertRaises(TypeError, s.add, {0: 1}, 0)
        self.assertEqual(s._table, {})

    def test_add_read_only(self):
        """
        Test that we cannot add when read-only (based on cache element).
        """
        s = MemoryKeyValueStore(DataMemoryElement(readonly=True))

        self.assertRaises(ReadOnlyError, s.add, 'a', 'b')
        self.assertRaises(ReadOnlyError, s.add, 'foo', None)

    def test_add(self):
        """ Test that we can add key-value pairs. """
        s = MemoryKeyValueStore()

        s.add('a', 'b')
        self.assertEqual(s._table, {'a': 'b'})

        s.add('foo', None)
        self.assertEqual(s._table, {
            'a': 'b',
            'foo': None,
        })

        s.add(0, 89)
        self.assertEqual(s._table, {
            'a': 'b',
            'foo': None,
            0: 89,
        })

    def test_add_with_caching(self):
        """
        Test that we can add key-value pairs and they reflect in the cache
        element.
        """
        c = DataMemoryElement()
        s = MemoryKeyValueStore(c)

        expected_cache_dict = {'a': 'b', 'foo': None, 0: 89}

        s.add('a', 'b')
        s.add('foo', None)
        s.add(0, 89)
        self.assertEqual(
            pickle.loads(c.get_bytes()),
            expected_cache_dict
        )

    def test_add_many_readonly(self):
        """ Test that we can't add many on a read-only instance. """
        s = MemoryKeyValueStore(DataMemoryElement(readonly=True))
        self.assertRaises(
            ReadOnlyError,
            s.add_many, {0: 1}
        )

    def test_add_many(self):
        """
        Test that we can add many key-values via a dictionary input.
        """
        d = {
            'a': 'b',
            'foo': None,
            0: 89,
        }

        s = MemoryKeyValueStore()
        self.assertIsNone(s._cache_element)
        self.assertEqual(s._table, {})

        s.add_many(d)
        self.assertIsNone(s._cache_element)
        self.assertEqual(s._table, d)

    def test_add_many_with_caching(self):
        """
        Test that adding many reflects in cache.
        """
        d = {
            'a': 'b',
            'foo': None,
            0: 89,
        }
        c = DataMemoryElement()

        s = MemoryKeyValueStore(c)
        self.assertEqual(s._table, {})
        self.assertEqual(c.get_bytes(), six.b(""))

        s.add_many(d)
        self.assertEqual(s._table, d)
        self.assertEqual(
            pickle.loads(c.get_bytes()),
            d
        )

    def test_remove_readonly(self):
        """ Test that we cannot remove from a read-only instance. """
        s = MemoryKeyValueStore(DataMemoryElement(readonly=True))
        self.assertRaises(
            ReadOnlyError,
            s.remove, 0
        )

    def test_remove_missing_key(self):
        """
        Test that we cannot remove a key not in the store.
        """
        s = MemoryKeyValueStore()
        s._table = {
            0: 1,
            'a': 'b'
        }
        self.assertRaises(
            KeyError,
            s.remove, 'some-key'
        )
        # table should remain unchanged.
        self.assertDictEqual(s._table, {0: 1, 'a': 'b'})

    def test_remove_with_cache(self):
        """
        Test that removal correctly updates the cache element.
        """
        existing_data = {
            0: 1,
            'a': 'b',
        }

        c = DataMemoryElement(pickle.dumps(existing_data))
        s = MemoryKeyValueStore(c)
        self.assertDictEqual(s._table, existing_data)

        s.remove('a')
        self.assertDictEqual(s._table, {0: 1})
        self.assertDictEqual(pickle.loads(c.get_bytes()),
                             {0: 1})

    def test_remove(self):
        """ Test normal removal. """
        s = MemoryKeyValueStore()
        s._table = {
            0: 1,
            'a': 'b',
        }

        s.remove(0)
        self.assertDictEqual(s._table, {'a': 'b'})

    def test_remove_many_readonly(self):
        """ Test that we cannot remove many from a read-only instance. """
        s = MemoryKeyValueStore(DataMemoryElement(readonly=True))
        self.assertRaises(
            ReadOnlyError,
            s.remove_many, [0]
        )

    def test_remove_many_missing_key(self):
        """
        Test that we cannot remove keys not present in table and that table
        is not modified on error.
        """
        expected_table = {
            0: 0,
            1: 1,
            2: 2,
        }

        s = MemoryKeyValueStore()
        s._table = {
            0: 0,
            1: 1,
            2: 2,
        }

        self.assertRaisesRegexp(
            KeyError, 'a',
            s.remove_many, ['a']
        )
        self.assertDictEqual(s._table, expected_table)

        # Even if one of the keys is value, the table should not be modified if
        # one of the keys is invalid.
        self.assertRaisesRegexp(
            KeyError, '6',
            s.remove_many, [1, 6]
        )
        self.assertDictEqual(s._table, expected_table)

        PY2_SET_KEY_ERROR_RE = "set\(\[(?:7|8), (?:7|8)\]\)"
        PY3_SET_KEY_ERROR_RE = "{(?:7|8), (?:7|8)}"
        self.assertRaisesRegexp(
            KeyError,
            # Should show a "set" that contains 7 and 8, regardless of order.
            '(?:{}|{})'.format(PY2_SET_KEY_ERROR_RE, PY3_SET_KEY_ERROR_RE),
            s.remove_many, [7, 8]
        )

    def test_remove_many(self):
        """
        Test expected remove_many functionality.
        """
        s = MemoryKeyValueStore()
        s._table = {
            0: 0,
            1: 1,
            2: 2,
        }

        s.remove_many([0, 1])
        self.assertDictEqual(s._table, {2: 2})

    def test_remove_many_with_cache(self):
        starting_table = {
            0: 0,
            1: 1,
            2: 2,
        }
        c = DataMemoryElement(pickle.dumps(starting_table))
        s = MemoryKeyValueStore(c)
        self.assertDictEqual(s._table, starting_table)

        s.remove_many([0, 2])

        self.assertDictEqual(pickle.loads(c.get_bytes()), {1: 1})

    def test_get_invalid_key(self):
        """
        Test that trying to get a value from a key not entered yields an
        exception.
        """
        s = MemoryKeyValueStore()
        self.assertRaises(
            KeyError,
            s.get, 0
        )

    def test_get_invalid_key_with_default(self):
        """ Test default value return on missing key. """
        s = MemoryKeyValueStore()
        self.assertEqual(
            s.get(0, 1),
            1,
        )
        assert s.get(0, ()) == ()

    def test_get_invalid_key_with_default_None(self):
        """ Test that passing None as a default returns None appropriately. """
        s = MemoryKeyValueStore()
        self.assertIsNone(s.get(0, None))

    def test_get(self):
        """ Test normal get functionality. """
        s = MemoryKeyValueStore()
        s._table['a'] = 'b'
        s._table[0] = 1

        assert s.get('a') == 'b'
        assert s.get(0) == 1

    def test_clear_readonly(self):
        """ Test trying to clear on a read-only store. """
        table_before_clear = dict(a=1, b=2, c=3)

        s = MemoryKeyValueStore()
        s._table = table_before_clear
        s.is_read_only = mock.MagicMock(return_value=True)

        self.assertRaises(
            ReadOnlyError,
            s.clear
        )
        self.assertEqual(s._table, table_before_clear)

    def test_clear(self):
        """ Test normal clear functionality. """
        table_before_clear = dict(a=1, b=2, c=3)

        s = MemoryKeyValueStore()
        s._table = table_before_clear
        s.clear()
        self.assertEqual(s._table, {})



"""
Train and generate models for the SMQTK IQR Application.

This application takes the same configuration file as the IqrService REST
service.  To generate a default configuration, please refer to the
``runApplication`` tool for the ``IqrService`` application:

    runApplication -a IqrService -g config.IqrService.json
"""
import argparse
import glob
import json
import logging
import os.path as osp

import six

from smqtk import algorithms
from smqtk import representation
from smqtk.representation.data_element.file_element import DataFileElement
from smqtk.utils import bin_utils, jsmin, plugin
from smqtk.web.iqr_service import IqrService


__author__ = 'paul.tunison@kitware.com'


def cli_parser():
    # Forgoing the ``bin_utils.basic_cli_parser`` due to our use of dual
    # configuration files for this utility.
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument('-v', '--verbose',
                        default=False, action='store_true',
                        help='Output additional debug logging.')
    parser.add_argument('-c', '--config',
                        metavar="PATH", nargs=2, required=True,
                        help='Path to the JSON configuration files. The first '
                             'file provided should be the configuration file '
                             'for the ``IqrSearchDispatcher`` web-application '
                             'and the second should be the configuration file '
                             'for the ``IqrService`` web-application.')

    parser.add_argument("-t", "--tab",
                        default=None, required=True,
                        help="The configuration \"tab\" of the "
                             "``IqrSearchDispatcher`` configuration to use. "
                             "This informs what dataset to add the input data "
                             "files to.")
    parser.add_argument("input_files",
                        metavar='GLOB', nargs="+",
                        help="Shell glob to files to add to the configured "
                             "data set.")

    return parser


def main():
    args = cli_parser().parse_args()

    ui_config_filepath, iqr_config_filepath = args.config
    llevel = logging.DEBUG if args.verbose else logging.INFO
    tab = args.tab
    input_files_globs = args.input_files

    # Not using `bin_utils.utility_main_helper`` due to deviating from single-
    # config-with-default usage.
    bin_utils.initialize_logging(logging.getLogger('smqtk'), llevel)
    bin_utils.initialize_logging(logging.getLogger('__main__'), llevel)
    log = logging.getLogger(__name__)

    log.info("Loading UI config: '{}'".format(ui_config_filepath))
    ui_config, ui_config_loaded = bin_utils.load_config(ui_config_filepath)
    log.info("Loading IQR config: '{}'".format(iqr_config_filepath))
    iqr_config, iqr_config_loaded = bin_utils.load_config(iqr_config_filepath)
    if not (ui_config_loaded and iqr_config_loaded):
        raise RuntimeError("One or both configuration files failed to load.")

    # Ensure the given "tab" exists in UI configuration.
    if tab is None:
        log.error("No configuration tab provided to drive model generation.")
        exit(1)
    if tab not in ui_config["iqr_tabs"]:
        log.error("Invalid tab provided: '{}'. Available tags: {}"
                  .format(tab, list(ui_config["iqr_tabs"])))
        exit(1)

    #
    # Gather Configurations
    #
    log.info("Extracting plugin configurations")

    ui_tab_config = ui_config["iqr_tabs"][tab]
    iqr_plugins_config = iqr_config['iqr_service']['plugins']

    # Configure DataSet implementation and parameters
    data_set_config = ui_tab_config['data_set']

    # Configure DescriptorElementFactory instance, which defines what
    # implementation of DescriptorElement to use for storing generated
    # descriptor vectors below.
    descriptor_elem_factory_config = iqr_plugins_config['descriptor_factory']

    # Configure DescriptorGenerator algorithm implementation, parameters and
    # persistent model component locations (if implementation has any).
    descriptor_generator_config = iqr_plugins_config['descriptor_generator']

    # Configure NearestNeighborIndex algorithm implementation, parameters and
    # persistent model component locations (if implementation has any).
    nn_index_config = iqr_plugins_config['neighbor_index']

    #
    # Initialize data/algorithms
    #
    # Constructing appropriate data structures and algorithms, needed for the
    # IQR demo application, in preparation for model training.
    #
    log.info("Instantiating plugins")
    #: :type: representation.DataSet
    data_set = \
        plugin.from_plugin_config(data_set_config,
                                  representation.get_data_set_impls())
    descriptor_elem_factory = \
        representation.DescriptorElementFactory \
        .from_config(descriptor_elem_factory_config)
    #: :type: algorithms.DescriptorGenerator
    descriptor_generator = \
        plugin.from_plugin_config(descriptor_generator_config,
                                  algorithms.get_descriptor_generator_impls())
    #: :type: algorithms.NearestNeighborsIndex
    nn_index = \
        plugin.from_plugin_config(nn_index_config,
                                  algorithms.get_nn_index_impls())

    #
    # Build models
    #
    log.info("Adding files to dataset '{}'".format(data_set))
    for g in input_files_globs:
        g = osp.expanduser(g)
        if osp.isfile(g):
            data_set.add_data(DataFileElement(g, readonly=True))
        else:
            log.debug("Expanding glob: %s" % g)
            for fp in glob.iglob(g):
                data_set.add_data(DataFileElement(fp, readonly=True))

    # Generate a model if the generator defines a known generation method.
    try:
        log.debug("descriptor generator as model to generate?")
        descriptor_generator.generate_model(data_set)
    except AttributeError as ex:
        log.debug("descriptor generator as model to generate - Nope: {}"
                  .format(str(ex)))

    # Generate descriptors of data for building NN index.
    log.info("Computing descriptors for data set with {}"
             .format(descriptor_generator))
    data2descriptor = descriptor_generator.compute_descriptor_async(
        data_set, descriptor_elem_factory
    )

    # Possible additional support steps before building NNIndex
    try:
        # Fit the LSH index functor
        log.debug("Has LSH Functor to fit?")
        nn_index.lsh_functor.fit(six.itervalues(data2descriptor))
    except AttributeError as ex:
        log.debug("Has LSH Functor to fit - Nope: {}".format(str(ex)))

    log.info("Building nearest neighbors index {}".format(nn_index))
    nn_index.build_index(six.itervalues(data2descriptor))


if __name__ == "__main__":
    main()



from smqtk.algorithms import Classifier
from smqtk.representation.data_element import from_uri


class IndexLabelClassifier (Classifier):
    """
    Applies a listing of labels (new-line separated) to input "descriptor"
    values, which is actually a vector of class confidence values.
    """

    @classmethod
    def is_usable(cls):
        return True

    def __init__(self, index_to_label_uri):
        """
        Construct a new "classifier" that applies labels to input vector
        indices.

        We expect to be given a URI to a new-line separated text file where each
        line is a separate label in order and matching the dimensionality of an
        input descriptor.

        :param index_to_label_uri: URI to new-line separated sequence of labels.
        :type index_to_label_uri: str

        """
        super(IndexLabelClassifier, self).__init__()

        # load label vector
        self.index_to_label_uri = index_to_label_uri
        self.label_vector = [l.strip() for l in
                             from_uri(index_to_label_uri).to_buffered_reader()]

    def get_config(self):
        """
        Return a JSON-compliant dictionary that could be passed to this class's
        ``from_config`` method to produce an instance with identical
        configuration.

        :return: JSON type compliant configuration dictionary.
        :rtype: dict

        """
        return {
            "index_to_label_uri": self.index_to_label_uri,
        }

    def get_labels(self):
        """
        Get a copy of the sequence of class labels that this classifier can
        classify descriptors into.

        :return: Sequence of possible classifier labels.
        :rtype: collections.Sequence[str]

        """
        # copying container
        return list(self.label_vector)

    def _classify(self, d):
        """
        Internal method that constructs the label-to-confidence map (dict) for
        a given DescriptorElement.

        The passed descriptor element is guaranteed to have a vector to extract.
        It is not extracted yet due to the philosophy of waiting until the
        vector is immediately needed. This moment is thus determined by the
        implementing algorithm.

        :param d: DescriptorElement containing the vector to classify.
        :type d: smqtk.representation.DescriptorElement

        :raises RuntimeError: Could not perform classification for some reason
            (see message in raised exception).

        :return: Dictionary mapping trained labels to classification confidence
            values.
        :rtype: dict[str, float]

        """
        d_vector = d.vector()
        if len(self.label_vector) != len(d_vector):
            raise RuntimeError("Failed to apply label vector to input "
                               "descriptor of incongruous dimensionality (%d "
                               "labels != %d vector dimension)"
                               % tuple(map(len, [self.label_vector, d_vector])))
        return dict(zip(self.label_vector, d_vector))



from __future__ import division, print_function
import unittest

import mock
import pytest

import smqtk.representation
import smqtk.exceptions


class DummyCEImpl (smqtk.representation.ClassificationElement):

    @classmethod
    def is_usable(cls):
        return True

    def get_config(self):
        return {}

    def get_classification(self):
        pass

    def set_classification(self, m=None, **kwds):
        return super(DummyCEImpl, self).set_classification(m, **kwds)

    def has_classifications(self):
        pass


class TestClassificationElementAbstract (unittest.TestCase):

    def test_init(self):
        e = DummyCEImpl('foo', 'bar')
        self.assertEqual(e.type_name, 'foo')
        self.assertEqual(e.uuid, 'bar')

    def test_hash(self):
        self.assertEqual(hash(DummyCEImpl('foo', 'bar')),
                         hash(('foo', 'bar')))

    def test_equality(self):
        e1 = DummyCEImpl('test', 0)
        e2 = DummyCEImpl('other', 1)
        e1.get_classification = e2.get_classification = \
            mock.Mock(return_value={1: 1, 2: 0})

        self.assertEqual(e1, e2)

    def test_not_equal(self):
        e1 = DummyCEImpl('test', 0)
        e1.get_classification = mock.Mock(return_value={1: 1, 2: 0})

        e2 = DummyCEImpl('other', 1)
        e2.get_classification = mock.Mock(return_value={1: 0, 2: 1})

        self.assertNotEqual(e1, e2)

    def test_get_items(self):
        e1 = DummyCEImpl('test', 0)
        e1.get_classification = mock.Mock(return_value={1: 1, 2: 0})

        # There should not be a KeyError for accessing the test labels, but
        # there should for a different label that is not includes in the element
        self.assertEqual(e1[1], 1)
        self.assertEqual(e1[2], 0)
        with pytest.raises(KeyError):
            # noinspection PyStatementEffect
            e1['some other key']

    def test_max_label(self):
        e = DummyCEImpl('test', 0)

        e.get_classification = mock.Mock(return_value={})
        self.assertRaises(
            smqtk.exceptions.NoClassificationError,
            e.max_label
        )

        e.get_classification = mock.Mock(return_value={1: 0, 2: 1})
        self.assertNotEqual(e.max_label(), 1)
        self.assertEqual(e.max_label(), 2)

    def test_set_no_input(self):
        e = DummyCEImpl('test', 0)
        self.assertRaises(
            ValueError,
            e.set_classification,
        )

    def test_set_empty_input(self):
        e = DummyCEImpl('test', 0)
        self.assertRaises(
            ValueError,
            e.set_classification,
            {}
        )

    def test_set_input_dict(self):
        e = DummyCEImpl('test', 0)
        v = {1: 0, 2: 1}
        self.assertEqual(e.set_classification(v),  v)

    def test_set_kwargs(self):
        e = DummyCEImpl('test', 0)
        self.assertEqual(e.set_classification(a=1, b=0),
                         {'a': 1, 'b': 0})

    def test_set_mixed(self):
        e = DummyCEImpl('test', 0)
        self.assertEqual(
            e.set_classification({'a': .25, 1: .25},
                                 b=.25, d=.25),
            {'a': .25, 1: .25, 'b': .25, 'd': .25}
        )

    def test_set_nonstandard(self):
        # Many classifiers output 1-sum confidence values, but not all (e.g.
        # CNN final layers like AlexNet).
        e = DummyCEImpl('test', 0)
        self.assertEqual(
            e.set_classification({'a': 1, 1: 1},
                                 b=1, d=1),
            {'a': 1, 1: 1, 'b': 1, 'd': 1}
        )


class TestClassificationElementAbstractImplGetter (unittest.TestCase):

    def test_get_plugins(self):
        # There are at least 2 internally provided implementations that will
        # always be available:
        #   MemoryClassificationElement
        #   FileClassificationElement
        m = smqtk.representation.get_classification_element_impls()
        assert len(m) >= 2
        self.assertIn('MemoryClassificationElement', m)
        self.assertIn('FileClassificationElement', m)



"""
Server for hosting proxy manager which hosts proxy object instances.

This takes a simple configuration file that looks like the following:

|   [server]
|   port = <integer>
|   authkey = <string>

"""
# Copyright 2013-2016 by Kitware, Inc. All Rights Reserved. Please refer to
# LICENSE.txt for licensing information, or contact General Counsel,
# Kitware, Inc., 28 Corporate Drive, Clifton Park, NY 12065.

from smqtk.utils import bin_utils
from smqtk.utils.proxy_manager import ProxyManager


def default_config():
    return {
        "port": 5000,
        "authkey": "CHANGE_ME",
    }


def cli_parser():
    return bin_utils.basic_cli_parser(__doc__)


def main():
    parser = cli_parser()
    args = parser.parse_args()

    # Default config options for this util are technically valid for running,
    # its just a bad authkey.
    config = bin_utils.utility_main_helper(default_config, args,
                                           default_config_valid=True)

    port = int(config['port'])
    authkey = str(config['authkey'])

    mgr = ProxyManager(('', port), authkey)
    mgr.get_server().serve_forever()


if __name__ == '__main__':
    main()



from __future__ import division
import logging

import numpy as np

from smqtk.algorithms.nn_index.lsh.functors import LshFunctor
from smqtk.representation.descriptor_element import elements_to_matrix
from smqtk.utils.bin_utils import report_progress


class SimpleRPFunctor (LshFunctor):
    """
    This class is meant purely as a baseline comparison for other
    LshFunctors and NNIndex plugins. It is not meant to be used in
    production, as it is unlikely to produce a quality index.
    """

    @classmethod
    def is_usable(cls):
        return True

    def __init__(self, bit_length=8, normalize=None, random_seed=None):
        super(SimpleRPFunctor, self).__init__()

        self.bit_length = bit_length
        self.normalize = normalize
        self.random_seed = random_seed

        # Model components
        self.rps = None
        self.mean_vec = None

    def _norm_vector(self, v):
        """
        Class standard array normalization. Normalized along max dimension
        (a=0 for a 1D array, a=1 for a 2D array, etc.).

        :param v: Vector to normalize
        :type v: numpy.ndarray

        :return: Returns the normalized version of input array ``v``.
        :rtype: numpy.ndarray

        """
        vm = v - self.mean_vec
        if self.normalize is None:
            # Normalization off
            return vm

        n = np.linalg.norm(vm, ord=self.normalize, axis=-1, keepdims=True)
        with np.errstate(divide='ignore'):
            return np.nan_to_num(vm/n)

    def get_config(self):
        return {
            "bit_length": self.bit_length,
            "normalize": self.normalize,
            "random_seed": self.random_seed,
        }

    def has_model(self):
        return self.mean_vec is not None

    def fit(self, descriptors, use_multiprocessing=True):
        """
        Fit the ITQ model given the input set of descriptors

        :param descriptors: Iterable of ``DescriptorElement`` vectors to fit
            the model to.
        :type descriptors:
            collections.Iterable[smqtk.representation.DescriptorElement]

        :param use_multiprocessing: If multiprocessing should be used, as
            opposed to threading, for collecting descriptor vectors from the
            provided iterable.
        :type use_multiprocessing: bool

        :raises RuntimeError: There is already a model loaded

        :return: Matrix hash codes for provided descriptors in order.
        :rtype: numpy.ndarray[bool]

        """
        if self.has_model():
            raise RuntimeError("Model components have already been loaded.")

        dbg_report_interval = None
        if self.get_logger().getEffectiveLevel() <= logging.DEBUG:
            dbg_report_interval = 1.0  # seconds
        if not hasattr(descriptors, "__len__"):
            self._log.info("Creating sequence from iterable")
            descriptors_l = []
            rs = [0]*7
            for d in descriptors:
                descriptors_l.append(d)
                report_progress(self._log.debug, rs, dbg_report_interval)
            descriptors = descriptors_l
        self._log.info("Creating matrix of descriptors for fitting")
        x = elements_to_matrix(
            descriptors, report_interval=dbg_report_interval,
            use_multiprocessing=use_multiprocessing)
        self._log.debug("descriptor matrix shape: %s", x.shape)
        n, dim = x.shape

        self._log.debug("Generating random projections")
        np.random.seed(self.random_seed)
        self.rps = np.random.randn(dim, self.bit_length)

        self._log.debug("Info normalizing descriptors with norm type: %s",
                        self.normalize)
        return self.get_hash(x)

    def get_hash(self, descriptor):
        if self.rps is None:
            raise RuntimeError("Random projection model not constructed. Call "
                               "`fit` first!")
        b = (self._norm_vector(descriptor).dot(self.rps) >= 0.0)
        return b.squeeze()



from __future__ import division, print_function
import unittest

import mock
import numpy
import six

from smqtk.representation.descriptor_element.local_elements import \
    DescriptorMemoryElement
from smqtk.algorithms.nn_index import NearestNeighborsIndex, get_nn_index_impls
from smqtk.utils.iter_validation import check_empty_iterable


class DummySI (NearestNeighborsIndex):

    @classmethod
    def is_usable(cls):
        """ stub """
        return True

    def get_config(self):
        """ stub """

    def _build_index(self, descriptors):
        """ stub """

    def _update_index(self, descriptors):
        """ stub """

    def _remove_from_index(self, uids):
        """ stub """

    def _nn(self, d, n=1):
        """ stub """

    def count(self):
        return 0


class TestNNIndexAbstract (unittest.TestCase):

    def test_get_impls(self):
        # Some implementations should be returned
        m = get_nn_index_impls()
        self.assertTrue(m)
        for cls in six.itervalues(m):
            self.assertTrue(issubclass(cls, NearestNeighborsIndex))

    def test_empty_iterable_exception(self):
        v = DummySI._empty_iterable_exception()
        self.assertIsInstance(v, ValueError)
        self.assertRegexpMatches(str(v), "DescriptorElement")

    def test_check_empty_iterable_no_data(self):
        # Test that an exception is thrown when an empty list/iterable is
        # passed.  Additionally check that the exception thrown has expected
        # message from exception generation method.
        callback = mock.MagicMock()

        # Not-stateful iterable (list)
        self.assertRaisesRegexp(
            ValueError,
            str(DummySI._empty_iterable_exception()),
            check_empty_iterable, [], callback,
            DummySI._empty_iterable_exception()
        )
        callback.assert_not_called()

        # with a stateful iterator.
        self.assertRaisesRegexp(
            ValueError,
            str(DummySI._empty_iterable_exception()),
            check_empty_iterable, iter([]), callback,
            DummySI._empty_iterable_exception()
        )
        callback.assert_not_called()

    def test_check_empty_iterable_valid_iterable(self):
        # Test that the method correctly calls the callback with the full
        # iterable when what is passed is not empty.
        callback = mock.MagicMock()

        # non-stateful iterator (set)
        d_set = {0, 1, 2, 3, 4}
        check_empty_iterable(d_set, callback,
                             DummySI()._empty_iterable_exception())
        callback.assert_called_once()
        self.assertSetEqual(
            set(callback.call_args[0][0]),
            d_set
        )

        # Stateful iterator
        callback = mock.MagicMock()
        check_empty_iterable(iter(d_set), callback,
                             DummySI()._empty_iterable_exception())
        callback.assert_called_once()
        self.assertSetEqual(
            set(callback.call_args[0][0]),
            d_set
        )

    def test_count_and_len(self):
        index = DummySI()
        self.assertEqual(index.count(), 0)
        self.assertEqual(index.count(), len(index))

        # Pretend that there were things in there. Len should pass it though
        index.count = mock.Mock()
        index.count.return_value = 5
        self.assertEqual(len(index), 5)

    def test_build_index_no_descriptors(self):
        index = DummySI()
        index._build_index = mock.MagicMock()
        self.assertRaises(
            ValueError,
            index.build_index,
            []
        )
        index._build_index.assert_not_called()

    def test_build_index_nonzero_descriptors(self):
        index = DummySI()
        index._build_index = mock.MagicMock()
        d = DescriptorMemoryElement('test', 0)
        index.build_index([d])
        index._build_index.assert_called_once()
        # Check that the last call's first (only) argument was the same iterable
        # given.
        self.assertSetEqual(
            set(index._build_index.call_args[0][0]),
            {d}
        )

    def test_build_index_iterable(self):
        # Test build check with a pure iterable
        index = DummySI()
        index._build_index = mock.MagicMock()
        d_set = {
            DescriptorMemoryElement('test', 0),
            DescriptorMemoryElement('test', 1),
            DescriptorMemoryElement('test', 2),
            DescriptorMemoryElement('test', 3),
        }
        it = iter(d_set)
        index.build_index(it)
        # _build_index should have been called and the contents of the iterable
        # it was called with should equal d_set.
        index._build_index.assert_called_once()
        self.assertSetEqual(
            set(index._build_index.call_args[0][0]),
            d_set
        )

    def test_update_index_no_descriptors(self):
        index = DummySI()
        index._update_index = mock.MagicMock()
        self.assertRaises(
            ValueError,
            index.update_index,
            []
        )
        # internal method should not have been called.
        index._update_index.assert_not_called()

    def test_update_index_nonzero_descriptors(self):
        index = DummySI()
        index._update_index = mock.MagicMock()

        # Testing with dummy input data.
        d_set = {
            DescriptorMemoryElement('test', 0),
            DescriptorMemoryElement('test', 1),
            DescriptorMemoryElement('test', 2),
            DescriptorMemoryElement('test', 3),
        }
        index.update_index(d_set)
        index._update_index.assert_called_once()
        self.assertSetEqual(
            set(index._update_index.call_args[0][0]),
            d_set
        )

    def test_update_index_iterable(self):
        # Test build check with a pure iterable.
        index = DummySI()
        index._update_index = mock.MagicMock()
        d_set = {
            DescriptorMemoryElement('test', 0),
            DescriptorMemoryElement('test', 1),
            DescriptorMemoryElement('test', 2),
            DescriptorMemoryElement('test', 3),
        }
        it = iter(d_set)
        index.update_index(it)

        index._update_index.assert_called_once()
        self.assertSetEqual(
            set(index._update_index.call_args[0][0]),
            d_set,
        )

    def test_remove_from_index_no_uids(self):
        # Test that the method errors when no UIDs are provided
        index = DummySI()
        index._remove_from_index = mock.Mock()
        self.assertRaises(
            ValueError,
            index.remove_from_index, []
        )
        index._remove_from_index.assert_not_called()

    def test_remove_from_index_nonzero_descriptors(self):
        # Test removing a non-zero amount of descriptors
        index = DummySI()
        index._remove_from_index = mock.MagicMock()

        # Testing with dummy input data.
        uid_set = {0, 1, 2, 3}
        index.remove_from_index(uid_set)
        index._remove_from_index.assert_called_once()
        self.assertSetEqual(
            set(index._remove_from_index.call_args[0][0]),
            uid_set
        )

    def test_remove_from_index_nonzero_iterable(self):
        # Test removing a non-zero amount of descriptors via an iterable.
        index = DummySI()
        index._remove_from_index = mock.MagicMock()
        d_set = {0, 1, 2, 3}
        it = iter(d_set)
        index.remove_from_index(it)

        index._remove_from_index.assert_called_once()
        self.assertSetEqual(
            set(index._remove_from_index.call_args[0][0]),
            d_set,
        )

    def test_nn_empty_vector(self):
        # ValueError should be thrown if the input element has no vector.
        index = DummySI()
        # Need to force a non-zero index size for knn to be performed.
        index.count = mock.MagicMock(return_value=1)
        # Observe internal function
        index._nn = mock.MagicMock()

        q = DescriptorMemoryElement('test', 0)
        self.assertRaises(
            ValueError,
            index.nn, q
        )
        # template method should not have been called.
        index._nn.assert_not_called()

    def test_nn_empty_index(self):
        # nn should fail if index size is 0
        index = DummySI()
        index.count = mock.MagicMock(return_value=0)
        index._nn = mock.MagicMock()

        q = DescriptorMemoryElement('q', 0)
        q.set_vector(numpy.random.rand(4))
        self.assertRaises(
            ValueError,
            index.nn, q
        )

    # noinspection PyUnresolvedReferences
    def test_nn_normal_conditions(self):
        index = DummySI()
        # Need to force a non-zero index size for knn to be performed.
        index.count = mock.MagicMock()
        index.count.return_value = 1

        q = DescriptorMemoryElement('q', 0)
        q.set_vector(numpy.random.rand(4))
        # Basically this shouldn't crash
        index.nn(q)

    def test_query_empty_index(self):
        index = DummySI()
        q = DescriptorMemoryElement('q', 0)
        q.set_vector(numpy.random.rand(4))
        self.assertRaises(ValueError, index.nn, q)



from __future__ import division, print_function
import abc
import logging
import json
import math
import mimetypes
import multiprocessing
import multiprocessing.pool
import os
import os.path as osp
import subprocess
import sys
import tempfile

import six
import numpy
import six
import sklearn.cluster

from smqtk.algorithms.descriptor_generator import DescriptorGenerator
from smqtk.representation.data_element.file_element import DataFileElement
from smqtk.utils import file_utils, SimpleTimer, video_utils
from smqtk.utils.string_utils import partition_string
from smqtk.utils.video_utils import get_metadata_info

# Attempt importing utilities module. If not, flag descriptor as unusable.
from . import utils

from six import next


# Requires FLANN bindings
try:
    import pyflann
except ImportError:
    pyflann = None


# noinspection PyPep8Naming
class ColorDescriptor_Base (DescriptorGenerator):
    """
    Simple implementation of ColorDescriptor feature descriptor utility for
    feature generation over images and videos.

    This was started as an attempt at gaining a deeper understanding of what's
    going on with this feature descriptor's use and how it applied to later use
    in an indexer.

    Codebook generated via kmeans given a set of input data. FLANN index model
    used for quantization, built using auto-tuning (picks the best indexing
    algorithm of linear, kdtree, kmeans, or combined), and using the Chi-Squared
    distance function.

    """

    # Name/Path to the colorDescriptor executable to use. By default we assume
    # its accessible on the PATH.
    EXE = 'colorDescriptor'

    @classmethod
    def is_usable(cls):
        """
        Check whether this descriptor is available for use.

        :return: Boolean determination of whether this implementation is usable.
        :rtype: bool

        """
        log = logging.getLogger('.'.join([ColorDescriptor_Base.__module__,
                                          ColorDescriptor_Base.__name__,
                                          "is_usable"]))

        if not hasattr(ColorDescriptor_Base, "_is_usable_cache"):
            # Base assumption that it is available, now lets prove its false.
            ColorDescriptor_Base._is_usable_cache = True

            if pyflann is None:
                # missing FLANN bindings dependency
                log.debug("could not import FLANN bindings (pyflann)")
                ColorDescriptor_Base._is_usable_cache = False
            else:
                # Check for colorDescriptor executable on the path
                log_file = open(tempfile.mkstemp()[1], 'w')
                try:
                    # This should try to print out the CLI options return with
                    # code 1.
                    subprocess.call([cls.EXE, '--version'],
                                    stdout=log_file, stderr=log_file)
                    # it is known that colorDescriptor has a return code of 1 no
                    # matter if it exited "successfully" or not, which is not
                    # helpful, I know.
                except OSError:
                    log.debug("Could not locate colorDescriptor executable. "
                              "Make sure that its on the PATH! See "
                              "smqtk/descriptor_generator/colordescriptor/"
                              "INSTALL.md for help.")
                    # If there was anything written to the log file, output it.
                    log_file.flush()
                    if log_file.tell():
                        with open(log_file.name) as f:
                            log.debug("STDOUT and STDERR output from attempted "
                                      "colorDescriptor call:\n%s", f.read())
                    ColorDescriptor_Base._is_usable_cache = False
                finally:
                    log_file.close()
                    os.remove(log_file.name)

                # Checking if DescriptorIO is importable
                if not utils.has_colordescriptor_module():
                    log.debug("Could not import DescriptorIO. Make sure that "
                              "the colorDescriptor package is on the "
                              "PYTHONPATH! See "
                              "smqtk/descriptor_generator/colordescriptor/"
                              "INSTALL.md for help.")
                    ColorDescriptor_Base._is_usable_cache = False

        return ColorDescriptor_Base._is_usable_cache

    def __init__(self, model_directory, work_directory,
                 model_gen_descriptor_limit=1000000,
                 kmeans_k=1024, flann_distance_metric='hik',
                 flann_target_precision=0.95,
                 flann_sample_fraction=0.75,
                 flann_autotune=False,
                 random_seed=None, use_spatial_pyramid=False,
                 parallel=None):
        """
        Initialize a new ColorDescriptor interface instance.

        :param model_directory: Path to the directory to store/read data model
            files on the local filesystem. Relative paths are treated relative
            to the current working directory.
        :type model_directory: str | unicode

        :param work_directory: Path to the directory in which to place
            temporary/working files. Relative paths are treated relative to
            the current working directory.
        :type work_directory: str | unicode

        :param model_gen_descriptor_limit: Total number of descriptors to use
            from input data to generate codebook model. Fewer than this may be
            used if the data set is small, but if it is greater, we randomly
            sample down to this count (occurs on a per element basis).
        :type model_gen_descriptor_limit: int

        :param kmeans_k: Centroids to generate. Default of 1024
        :type kmeans_k: int

        :param flann_distance_metric: Distance function to use in FLANN
            indexing. See FLANN documentation for available distance function
            types (under the MATLAB section reference for valid string
            identifiers)
        :type flann_distance_metric: str

        :param flann_target_precision: Target precision percent to tune index
            for. Default is 0.95 (95% accuracy). For some codebooks, if this is
            too close to 1.0, the FLANN library may non-deterministically
            overflow, causing an infinite loop requiring a SIGKILL to stop.
        :type flann_target_precision: float

        :param flann_sample_fraction: Fraction of input data to use for index
            auto tuning. Default is 0.75 (75%).
        :type flann_sample_fraction: float

        :param flann_autotune: Have FLANN module use auto-tuning algorithm to
            find an optimal index representation and parameter set.
        :type flann_autotune: bool

        :param random_seed: Optional value to seed components requiring random
            operations.
        :type random_seed: None or int

        :param use_spatial_pyramid: Use spacial pyramids when quantizing low
            level descriptors during feature computation.
        :type use_spatial_pyramid: bool

        :param parallel: Specific number of threads/cores to use when performing
            asynchronous activities. When None we will use all cores available.
        :type parallel: int | None

        """
        super(ColorDescriptor_Base, self).__init__()

        # TODO: Because of the FLANN library non-deterministic overflow issue,
        #       an alternative must be found before this can be put into
        #       production. Suggest saving/using sk-learn MBKMeans class? Can
        #       the class be regenerated from an existing codebook?
        self._model_dir = model_directory
        self._work_dir = work_directory

        self._model_gen_descriptor_limit = model_gen_descriptor_limit

        self._kmeans_k = int(kmeans_k)
        self._flann_distance_metric = flann_distance_metric
        self._flann_target_precision = float(flann_target_precision)
        self._flann_sample_fraction = float(flann_sample_fraction)
        self._flann_autotune = bool(flann_autotune)
        self._use_sp = use_spatial_pyramid
        self._rand_seed = None if random_seed is None else int(random_seed)

        if self._rand_seed is not None:
            numpy.random.seed(self._rand_seed)

        self.parallel = parallel

        # Cannot pre-load FLANN stuff because odd things happen when processing/
        # threading. Loading index file is fast anyway.
        self._codebook = None
        if self.has_model:
            self._codebook = numpy.load(self.codebook_filepath)

    def get_config(self):
        """
        Return a JSON-compliant dictionary that could be passed to this class's
        ``from_config`` method to produce an instance with identical
        configuration.

        Keys of the returned dictionary are based on the initialization argument
        names.

        :return: JSON type compliant configuration dictionary.
        :rtype: dict

        """
        return {
            "model_directory": self._model_dir,
            "work_directory": self._work_dir,
            "model_gen_descriptor_limit": self._model_gen_descriptor_limit,
            "kmeans_k": self._kmeans_k,
            "flann_distance_metric": self._flann_distance_metric,
            "flann_target_precision": self._flann_target_precision,
            "flann_sample_fraction": self._flann_sample_fraction,
            "flann_autotune": self._flann_autotune,
            "random_seed": self._rand_seed,
            "use_spatial_pyramid": self._use_sp,
        }

    @property
    def name(self):
        if self._use_sp:
            return '_'.join([self.__class__.__name__, 'spatial'])
        else:
            return self.__class__.__name__

    @property
    def codebook_filepath(self):
        file_utils.safe_create_dir(self._model_dir)
        return osp.join(self._model_dir,
                        "%s.codebook.npy" % (self.descriptor_type(),))

    @property
    def flann_index_filepath(self):
        file_utils.safe_create_dir(self._model_dir)
        return osp.join(self._model_dir,
                        "%s.flann_index.dat" % (self.descriptor_type(),))

    @property
    def flann_params_filepath(self):
        file_utils.safe_create_dir(self._model_dir)
        return osp.join(self._model_dir,
                        "%s.flann_params.json" % (self.descriptor_type(),))

    @property
    def has_model(self):
        has_model = (osp.isfile(self.codebook_filepath) and
                     osp.isfile(self.flann_index_filepath))
        # Load the codebook model if not already loaded. FLANN index will be
        # loaded when needed to prevent thread/subprocess memory issues.
        if self._codebook is None and has_model:
            self._codebook = numpy.load(self.codebook_filepath)
        return has_model

    @property
    def temp_dir(self):
        return file_utils.safe_create_dir(osp.join(self._work_dir,
                                                   'temp_files'))

    @abc.abstractmethod
    def descriptor_type(self):
        """
        :return: String descriptor type as used by colorDescriptor
        :rtype: str
        """
        return

    @abc.abstractmethod
    def _generate_descriptor_matrices(self, data_set, **kwargs):
        """
        Generate info and descriptor matrices based on ingest type.

        :param data_set: Iterable of data elements to generate combined info
            and descriptor matrices for.
        :type item_iter: collections.Set[smqtk.representation.DataElement]

        :param limit: Limit the number of descriptor entries to this amount.
        :type limit: int

        :return: Combined info and descriptor matrices for all base images
        :rtype: (numpy.core.multiarray.ndarray, numpy.core.multiarray.ndarray)

        """
        pass

    def _get_checkpoint_dir(self, data):
        """
        The directory that contains checkpoint material for a given data element

        :param data: Data element
        :type data: smqtk.representation.DataElement

        :return: directory path
        :rtype: str

        """
        d = osp.join(self._work_dir, *partition_string(str(data.uuid()), 10))
        file_utils.safe_create_dir(d)
        return d

    def _get_standard_info_descriptors_filepath(self, data, frame=None):
        """
        Get the standard path to a data element's computed descriptor output,
        which for colorDescriptor consists of two matrices: info and descriptors

        :param data: Data element
        :type data: smqtk.representation.DataElement

        :param frame: frame within the data file
        :type frame: int

        :return: Paths to info and descriptor checkpoint numpy files
        :rtype: (str, str)

        """
        d = self._get_checkpoint_dir(data)
        if frame is not None:
            return (
                osp.join(d, "%s.info.%d.npy" % (str(data.uuid()), frame)),
                osp.join(d, "%s.descriptors.%d.npy" % (str(data.uuid()), frame))
            )
        else:
            return (
                osp.join(d, "%s.info.npy" % str(data.uuid())),
                osp.join(d, "%s.descriptors.npy" % str(data.uuid()))
            )

    def _get_checkpoint_feature_file(self, data):
        """
        Return the standard path to a data element's computed feature checkpoint
        file relative to our current working directory.

        :param data: Data element
        :type data: smqtk.representation.DataElement

        :return: Standard path to where the feature checkpoint file for this
            given data element.
        :rtype: str

        """
        if self._use_sp:
            return osp.join(self._get_checkpoint_dir(data),
                            "%s.feature.sp.npy" % str(data.uuid()))
        else:
            return osp.join(self._get_checkpoint_dir(data),
                            "%s.feature.npy" % str(data.uuid()))

    def generate_model(self, data_set):
        """
        Generate this feature detector's data-model given a file ingest. This
        saves the generated model to the currently configured data directory.

        For colorDescriptor, we generate raw features over the ingest data,
        compute a codebook via kmeans, and then create an index with FLANN via
        the "autotune" or linear algorithm to intelligently pick the fastest
        indexing method.

        :param data_set: Set of input data elements to generate the model
            with.
        :type data_set: collections.Set[smqtk.representation.DataElement]

        """
        if self.has_model:
            self._log.warn("ColorDescriptor model for descriptor type '%s' "
                           "already generated!", self.descriptor_type())
            return

        # Check that input data is value for processing through colorDescriptor
        valid_types = self.valid_content_types()
        invalid_types_found = set()
        for di in data_set:
            if di.content_type() not in valid_types:
                invalid_types_found.add(di.content_type())
        if invalid_types_found:
            self._log.error("Found one or more invalid content types among "
                            "input:")
            for t in sorted(invalid_types_found):
                self._log.error("\t- '%s", t)
            raise ValueError("Discovered invalid content type among input "
                             "data: %s" % sorted(invalid_types_found))

        if not osp.isfile(self.codebook_filepath):
            self._log.info("Did not find existing ColorDescriptor codebook for "
                           "descriptor '%s'.", self.descriptor_type())

            # generate descriptors
            with SimpleTimer("Generating descriptor matrices...",
                             self._log.info):
                descriptors_checkpoint = osp.join(self._work_dir,
                                                  "model_descriptors.npy")

                if osp.isfile(descriptors_checkpoint):
                    self._log.debug("Found existing computed descriptors work "
                                    "file for model generation.")
                    descriptors = numpy.load(descriptors_checkpoint)
                else:
                    self._log.debug("Computing model descriptors")
                    _, descriptors = \
                        self._generate_descriptor_matrices(
                            data_set,
                            limit=self._model_gen_descriptor_limit
                        )
                    _, tmp = tempfile.mkstemp(dir=self._work_dir,
                                              suffix='.npy')
                    self._log.debug("Saving model-gen info/descriptor matrix")
                    numpy.save(tmp, descriptors)
                    os.rename(tmp, descriptors_checkpoint)

            # Compute centroids (codebook) with kmeans
            with SimpleTimer("Computing sklearn.cluster.MiniBatchKMeans...",
                             self._log.info):
                kmeans_verbose = self._log.getEffectiveLevel <= logging.DEBUG
                kmeans = sklearn.cluster.MiniBatchKMeans(
                    n_clusters=self._kmeans_k,
                    init_size=self._kmeans_k*3,
                    random_state=self._rand_seed,
                    verbose=kmeans_verbose,
                    compute_labels=False,
                )
                kmeans.fit(descriptors)
                codebook = kmeans.cluster_centers_
            with SimpleTimer("Saving generated codebook...", self._log.debug):
                numpy.save(self.codebook_filepath, codebook)
        else:
            self._log.info("Found existing codebook file.")
            codebook = numpy.load(self.codebook_filepath)

        # create FLANN index
        # - autotune will force select linear search if there are < 1000 words
        #   in the codebook vocabulary.
        pyflann.set_distance_type(self._flann_distance_metric)
        flann = pyflann.FLANN()
        if self._log.getEffectiveLevel() <= logging.DEBUG:
            log_level = 'info'
        else:
            log_level = 'warning'
        with SimpleTimer("Building FLANN index...", self._log.info):
            p = {
                "target_precision": self._flann_target_precision,
                "sample_fraction": self._flann_sample_fraction,
                "log_level": log_level,
            }
            if self._flann_autotune:
                p['algorithm'] = "autotuned"
            if self._rand_seed is not None:
                p['random_seed'] = self._rand_seed
            flann_params = flann.build_index(codebook, **p)
        with SimpleTimer("Saving FLANN index to file...", self._log.debug):
            # Save FLANN index data binary
            flann.save_index(self.flann_index_filepath)
            # Save out log of parameters
            with open(self.flann_params_filepath, 'w') as ofile:
                json.dump(flann_params, ofile, indent=4, sort_keys=True)

        # save generation results to class for immediate feature computation use
        self._codebook = codebook

    def _compute_descriptor(self, data):
        """
        Given some kind of data, process and return a feature vector as a Numpy
        array.

        :raises RuntimeError: Feature extraction failure of some kind.

        :param data: Some kind of input data for the feature descriptor. This is
            descriptor dependent.
        :type data: smqtk.representation.DataElement

        :return: Feature vector. This is a histogram of N bins where N is the
            number of centroids in the codebook. Bin values is percent
            composition, not absolute counts.
        :rtype: numpy.ndarray

        """
        super(ColorDescriptor_Base, self)._compute_descriptor(data)

        checkpoint_filepath = self._get_checkpoint_feature_file(data)
        # if osp.isfile(checkpoint_filepath):
        #     return numpy.load(checkpoint_filepath)

        if not self.has_model:
            raise RuntimeError("No model currently loaded! Check the existence "
                               "or, or generate, model files!\n"
                               "Codebook path: %s\n"
                               "FLANN Index path: %s"
                               % (self.codebook_filepath,
                                  self.flann_index_filepath))

        self._log.debug("Computing descriptors for data UID[%s]...",
                        data.uuid())
        info, descriptors = self._generate_descriptor_matrices({data})

        # Load FLANN components
        pyflann.set_distance_type(self._flann_distance_metric)
        flann = pyflann.FLANN()
        flann.load_index(self.flann_index_filepath, self._codebook)

        if not self._use_sp:
            ###
            # Codebook Quantization
            #
            # - loaded the model at class initialization if we had one
            self._log.debug("Quantizing descriptors")

            try:
                # If the distance method is HIK, we need to treat it special
                # since that method produces a similarity score, not a distance
                # score.
                #
                if self._flann_distance_metric == 'hik':
                    # This searches for all NN instead of minimum between n and
                    # the number of descriptors and keeps the last one because
                    # hik is a similarity score and not a distance, which is
                    # also why the values in dists is flipped below.
                    #: :type: (numpy.ndarray, numpy.ndarray)
                    idxs = flann.nn_index(descriptors,
                                          self._codebook.shape[0])[0]
                    # Only keep the last index for each descriptor return
                    idxs = numpy.array([i_array[-1] for i_array in idxs])
                else:
                    #: :type: (numpy.ndarray, numpy.ndarray)
                    idxs = flann.nn_index(descriptors, 1)[0]
            except AssertionError:

                self._log.error("Codebook shape  : %s", self._codebook.shape)
                self._log.error("Descriptor shape: %s", descriptors.shape)

                raise

            # Create histogram
            # - Using explicit bin slots to prevent numpy from automatically
            #   creating tightly constrained bins. This would otherwise cause
            #   histograms between two inputs to be non-comparable (unaligned
            #   bins).
            # - See numpy note about ``bins`` to understand why the +1 is
            #   necessary
            # - Learned from spatial implementation that we could feed multiple
            #   neighbors per descriptor into here, leading to a more populated
            #   histogram.
            #   - Could also possibly weight things based on dist from
            #     descriptor?
            #: :type: numpy.core.multiarray.ndarray
            h = numpy.histogram(idxs,  # indices are all integers
                                bins=numpy.arange(self._codebook.shape[0]+1))[0]
            # self._log.debug("Quantization histogram: %s", h)
            # Normalize histogram into relative frequencies
            # - Not using /= on purpose. h is originally int32 coming out of
            #   histogram. /= would keep int32 type when we want it to be
            #   transformed into a float type by the division.
            if h.sum():
                # noinspection PyAugmentAssignment
                h = h / float(h.sum())
            else:
                h = numpy.zeros(h.shape, h.dtype)
            # self._log.debug("Normalized histogram: %s", h)

        else:
            ###
            # Spatial Pyramid Quantization
            #
            self._log.debug("Quantizing descriptors using spatial pyramid")
            ##
            # Quantization factor - number of nearest codes to be saved
            q_factor = 10
            ##
            # Concatenating spatial information to descriptor vectors to format:
            #   [ x y <descriptor> ]
            self._log.debug("Creating combined descriptor matrix")
            m = numpy.concatenate((info[:, :2],
                                   descriptors), axis=1)
            ##
            # Creating quantized vectors, consisting vector:
            #   [ x y c_1 ... c_qf dist_1 ... dist_qf ]
            # which has a total size of 2+(qf*2)
            #
            # Sangmin's code included the distances in the quantized vector, but
            # then also passed this vector into numpy's histogram function with
            # integral bins, causing the [0,1] to be heavily populated, which
            # doesn't make sense to do.
            #   idxs, dists = flann.nn_index(m[:, 2:], q_factor)
            #   q = numpy.concatenate([m[:, :2], idxs, dists], axis=1)
            self._log.debug("Computing nearest neighbors")
            if self._flann_distance_metric == 'hik':
                # Query full ordering of code indices
                idxs = flann.nn_index(m[:, 2:], self._codebook.shape[0])[0]
                # Extract the right-side block for use in building histogram
                # Order doesn't actually matter in the current implementation
                #   because index relative position is not being weighted.
                idxs = idxs[:, -q_factor:]
            else:
                idxs = flann.nn_index(m[:, 2:], q_factor)[0]
            self._log.debug("Creating quantization matrix")
            # This matrix consists of descriptor (x,y) position + near code
            #   indices.
            q = numpy.concatenate([m[:, :2], idxs], axis=1)
            ##
            # Build spatial pyramid from quantized matrix
            self._log.debug("Building spatial pyramid histograms")
            hist_sp = self._build_sp_hist(q, self._codebook.shape[0])
            ##
            # Combine each quadrants into single vector
            self._log.debug("Combining global+thirds into final histogram.")
            f = sys.float_info.min  # so as we don't div by 0 accidentally

            def rf_norm(hist):
                return hist / (float(hist.sum()) + f)
            h = numpy.concatenate([rf_norm(hist_sp[0]),
                                   rf_norm(hist_sp[5]),
                                   rf_norm(hist_sp[6]),
                                   rf_norm(hist_sp[7])],
                                  axis=1)
            # noinspection PyAugmentAssignment
            h /= h.sum()

        self._log.debug("Saving checkpoint feature file")
        if not osp.isdir(osp.dirname(checkpoint_filepath)):
            file_utils.safe_create_dir(osp.dirname(checkpoint_filepath))
        numpy.save(checkpoint_filepath, h)

        return h

    @staticmethod
    def _build_sp_hist(feas, bins):
        """
        Build spatial pyramid from quantized data. We expect feature matrix
        to be in the following format:

            [[ x y c_1 ... c_n ]
             [ ... ]
             ... ]

        NOTES:
            - See encode_FLANN.py for original implementation this was adapted
                from.

        :param feas: Feature matrix with the above format.
        :type feas: numpy.core.multiarray.ndarray

        :param bins: number of bins for the spatial histograms. This should
            probably be the size of the codebook used when generating quantized
            descriptors.
        :type bins: int

        :return: Matrix of 8 rows representing the histograms for the different
            spatial regions
        :rtype: numpy.core.multiarray.ndarray

        """
        bins = numpy.arange(0, bins+1)
        cordx = feas[:, 0]
        cordy = feas[:, 1]
        feas = feas[:, 2:]

        # hard quantization
        # global histogram
        #: :type: numpy.core.multiarray.ndarray
        hist_sp_g = numpy.histogram(feas, bins=bins)[0]
        hist_sp_g = hist_sp_g[numpy.newaxis]
        # 4 quadrants
        # noinspection PyTypeChecker
        midx = numpy.ceil(cordx.max()/2)
        # noinspection PyTypeChecker
        midy = numpy.ceil(cordy.max()/2)
        lx = cordx < midx
        rx = cordx >= midx
        uy = cordy < midy
        dy = cordy >= midy
        # logging.error("LXUI: %s,%s", lx.__repr__(), uy.__repr__())
        # logging.error("Length LXUI: %s,%s", lx.shape, uy.shape)
        # logging.error("feas dimensions: %s", feas.shape)

        #: :type: numpy.core.multiarray.ndarray
        hist_sp_q1 = numpy.histogram(feas[lx & uy], bins=bins)[0]
        #: :type: numpy.core.multiarray.ndarray
        hist_sp_q2 = numpy.histogram(feas[rx & uy], bins=bins)[0]
        #: :type: numpy.core.multiarray.ndarray
        hist_sp_q3 = numpy.histogram(feas[lx & dy], bins=bins)[0]
        #: :type: numpy.core.multiarray.ndarray
        hist_sp_q4 = numpy.histogram(feas[rx & dy], bins=bins)[0]
        hist_sp_q1 = hist_sp_q1[numpy.newaxis]
        hist_sp_q2 = hist_sp_q2[numpy.newaxis]
        hist_sp_q3 = hist_sp_q3[numpy.newaxis]
        hist_sp_q4 = hist_sp_q4[numpy.newaxis]

        # 3 layers
        # noinspection PyTypeChecker
        ythird = numpy.ceil(cordy.max()/3)
        l1 = cordy <= ythird
        l2 = (cordy > ythird) & (cordy <= 2*ythird)
        l3 = cordy > 2*ythird
        #: :type: numpy.core.multiarray.ndarray
        hist_sp_l1 = numpy.histogram(feas[l1], bins=bins)[0]
        #: :type: numpy.core.multiarray.ndarray
        hist_sp_l2 = numpy.histogram(feas[l2], bins=bins)[0]
        #: :type: numpy.core.multiarray.ndarray
        hist_sp_l3 = numpy.histogram(feas[l3], bins=bins)[0]
        hist_sp_l1 = hist_sp_l1[numpy.newaxis]
        hist_sp_l2 = hist_sp_l2[numpy.newaxis]
        hist_sp_l3 = hist_sp_l3[numpy.newaxis]
        # concatenate
        hist_sp = numpy.vstack((hist_sp_g, hist_sp_q1, hist_sp_q2,
                                hist_sp_q3, hist_sp_q4, hist_sp_l1,
                                hist_sp_l2, hist_sp_l3))
        return hist_sp

    def _get_data_temp_path(self, de):
        """
        Standard method of generating/getting a data element's temporary file
        path.

        :param de: DataElement instance to generate/get temporary file path.
        :type de: smqtk.representation.DataElement

        :return: Path to the element's temporary file.
        :rtype: str

        """
        temp_dir = None
        # Shortcut when data element is from file, since we are not going to
        # write to / modify the file.
        if not isinstance(de, DataFileElement):
            temp_dir = self.temp_dir
        return de.write_temp(temp_dir)


# noinspection PyAbstractClass,PyPep8Naming
class ColorDescriptor_Image (ColorDescriptor_Base):

    def valid_content_types(self):
        """
        :return: A set valid MIME type content types that this descriptor can
            handle.
        :rtype: set[str]
        """
        return {'image/bmp', 'image/jpeg', 'image/png', 'image/tiff'}

    def _generate_descriptor_matrices(self, data_set, **kwargs):
        """
        Generate info and descriptor matrices based on ingest type.

        :param data_set: Iterable of data elements to generate combined info
            and descriptor matrices for.
        :type item_iter: collections.Set[smqtk.representation.DataElement]

        :param limit: Limit the number of descriptor entries to this amount.
        :type limit: int

        :return: Combined info and descriptor matrices for all base images
        :rtype: (numpy.core.multiarray.ndarray, numpy.core.multiarray.ndarray)

        """
        if not data_set:
            raise ValueError("No data given to process.")

        inf = float('inf')
        descriptor_limit = kwargs.get('limit', inf)
        per_item_limit = numpy.floor(float(descriptor_limit) / len(data_set))

        if len(data_set) == 1:
            # because an iterable doesn't necessarily have a next() method
            di = next(iter(data_set))
            # Check for checkpoint files
            info_fp, desc_fp = \
                self._get_standard_info_descriptors_filepath(di)
            # Save out data bytes to temporary file
            temp_img_filepath = self._get_data_temp_path(di)
            try:
                # Generate descriptors
                utils.generate_descriptors(
                    self.EXE, temp_img_filepath,
                    self.descriptor_type(), info_fp, desc_fp, per_item_limit
                )
            finally:
                # clean temp file
                di.clean_temp()
            return numpy.load(info_fp), numpy.load(desc_fp)
        else:
            # compute and V-stack matrices for all given images
            pool = multiprocessing.Pool(processes=self.parallel)

            # Mapping of UID to tuple containing:
            #   (info_fp, desc_fp, async processing result, tmp_clean_method)
            r_map = {}
            with SimpleTimer("Computing descriptors async...", self._log.debug):
                for di in data_set:
                    # Creating temporary image file from data bytes
                    tmp_img_fp = self._get_data_temp_path(di)

                    info_fp, desc_fp = \
                        self._get_standard_info_descriptors_filepath(di)
                    args = (self.EXE, tmp_img_fp,
                            self.descriptor_type(), info_fp, desc_fp)
                    r = pool.apply_async(utils.generate_descriptors, args)
                    r_map[di.uuid()] = (info_fp, desc_fp, r, di.clean_temp)
            pool.close()

            # Pass through results from descriptor generation, aggregating
            # matrix shapes.
            # - Transforms r_map into:
            #       UID -> (info_fp, desc_fp, starting_row, SubSampleIndices)
            self._log.debug("Constructing information for super matrices...")
            s_keys = sorted(r_map.keys())
            running_height = 0  # info and desc heights congruent

            i_width = None
            d_width = None

            for uid in s_keys:
                ifp, dfp, r, tmp_clean_method = r_map[uid]

                # descriptor generation may have failed for this ingest UID
                try:
                    i_shape, d_shape = r.get()
                except RuntimeError as ex:
                    self._log.warning("Descriptor generation failed for "
                                      "UID[%s], skipping its inclusion in "
                                      "model: %s", uid, str(ex))
                    r_map[uid] = None
                    continue
                finally:
                    # Done with image file, so remove from filesystem
                    tmp_clean_method()

                if d_width is None and d_shape[0] != 0:
                    i_width = i_shape[1]
                    d_width = d_shape[1]

                # skip this result if it generated no descriptors
                if d_shape[1] == 0:
                    continue

                ssi = None
                if i_shape[0] > per_item_limit:
                    # pick random indices to subsample down to size limit
                    ssi = sorted(
                        numpy.random.permutation(i_shape[0])[:per_item_limit]
                    )

                # Only keep this if any descriptors were generated
                r_map[uid] = (ifp, dfp, running_height, ssi)
                running_height += min(i_shape[0], per_item_limit)
            pool.join()

            # Asynchronously load files, inserting data into master matrices
            self._log.debug("Building super matrices...")
            master_info = numpy.zeros((running_height, i_width), dtype=float)
            master_desc = numpy.zeros((running_height, d_width), dtype=float)
            tp = multiprocessing.pool.ThreadPool(processes=self.parallel)
            for uid in s_keys:
                if r_map[uid]:
                    ifp, dfp, sR, ssi = r_map[uid]
                    tp.apply_async(ColorDescriptor_Image._thread_load_matrix,
                                   args=(ifp, master_info, sR, ssi))
                    tp.apply_async(ColorDescriptor_Image._thread_load_matrix,
                                   args=(dfp, master_desc, sR, ssi))
            tp.close()
            tp.join()
            return master_info, master_desc

    @staticmethod
    def _thread_load_matrix(filepath, m, sR, subsample=None):
        """
        load a numpy matrix from ``filepath``, inserting the loaded matrix into
        ``m`` starting at the row ``sR``.

        If subsample has a value, it will be a list if indices to
        """
        n = numpy.load(filepath)
        if subsample:
            n = n[subsample, :]
        m[sR:sR+n.shape[0], :n.shape[1]] = n


# noinspection PyAbstractClass,PyPep8Naming
class ColorDescriptor_Video (ColorDescriptor_Base):

    # # Custom higher limit for video since, ya know, they have multiple frames.
    CODEBOOK_DESCRIPTOR_LIMIT = 1500000

    FRAME_EXTRACTION_PARAMS = {
        "second_offset": 0.0,       # Start at beginning
        "second_interval": 0.5,     # Sample every 0.5 seconds
        "max_duration": 1.0,        # Cover full duration
        "output_image_ext": 'png',  # Output PNG files
        "ffmpeg_exe": "ffmpeg",
    }

    def valid_content_types(self):
        """
        :return: A set valid MIME type content types that this descriptor can
            handle.
        :rtype: set[str]
        """
        # At the moment, assuming ffmpeg can decode all video types, which it
        # probably cannot, but we'll filter this down when it becomes relevant.
        # noinspection PyUnresolvedReferences
        # TODO: GIF support?
        return set([x for x in mimetypes.types_map.values()
                    if x.startswith('video')])

    def _generate_descriptor_matrices(self, data_set, **kwargs):
        """
        Generate info and descriptor matrices based on ingest type.

        :param data_set: Iterable of data elements to generate combined info
            and descriptor matrices for.
        :type item_iter: collections.Set[smqtk.representation.DataElement]

        :param limit: Limit the number of descriptor entries to this amount.
        :type limit: int

        :return: Combined info and descriptor matrices for all base images
        :rtype: (numpy.core.multiarray.ndarray, numpy.core.multiarray.ndarray)

        """
        descriptor_limit = kwargs.get('limit', float('inf'))
        # With videos, an "item" is one video, so, collect for a while video
        # as normal, then subsample from the full video collection.
        per_item_limit = numpy.floor(float(descriptor_limit) / len(data_set))

        # If an odd number of jobs, favor descriptor extraction
        if self.parallel:
            descr_parallel = int(max(1.0, math.ceil(self.parallel/2.0)))
            extract_parallel = int(max(1.0, math.floor(self.parallel/2.0)))
        else:
            cpuc = multiprocessing.cpu_count()
            descr_parallel = int(max(1.0, math.ceil(cpuc/2.0)))
            extract_parallel = int(max(1.0, math.floor(cpuc/2.0)))

        # For each video, extract frames and submit colorDescriptor processing
        # jobs for each frame, combining all results into a single matrix for
        # return.
        pool = multiprocessing.Pool(processes=descr_parallel)

        # Mapping of [UID] to [frame] to tuple containing:
        #   (info_fp, desc_fp, async processing result)
        r_map = {}
        with SimpleTimer("Extracting frames and submitting descriptor jobs...",
                         self._log.debug):
            for di in data_set:
                r_map[di.uuid()] = {}
                tmp_vid_fp = self._get_data_temp_path(di)
                p = dict(self.FRAME_EXTRACTION_PARAMS)
                vmd = get_metadata_info(tmp_vid_fp)
                p['second_offset'] = vmd.duration * p['second_offset']
                p['max_duration'] = vmd.duration * p['max_duration']
                fm = video_utils.ffmpeg_extract_frame_map(
                    self._work_dir,
                    tmp_vid_fp,
                    parallel=extract_parallel,
                    **p
                )

                # Compute descriptors for extracted frames.
                for frame, imgPath in six.iteritems(fm):
                    info_fp, desc_fp = \
                        self._get_standard_info_descriptors_filepath(di, frame)
                    r = pool.apply_async(
                        utils.generate_descriptors,
                        args=(self.EXE, imgPath,
                              self.descriptor_type(), info_fp, desc_fp)
                    )
                    r_map[di.uuid()][frame] = (info_fp, desc_fp, r)

                # Clean temporary video file file while computing descriptors
                # This does not remove the extracted frames that the underlying
                #   detector/descriptor is working on.
                di.clean_temp()
        pool.close()

        # Each result is a tuple of two ndarrays: info and descriptor matrices
        with SimpleTimer("Collecting shape information for super matrices...",
                         self._log.debug):
            running_height = 0

            i_width = None
            d_width = None

            # Transform r_map[uid] into:
            #   (info_mat_files, desc_mat_files, sR, ssi_list)
            #   -> files in frame order
            uids = sorted(r_map)
            for uid in uids:
                video_num_desc = 0
                video_info_mat_fps = []  # ordered list of frame info mat files
                video_desc_mat_fps = []  # ordered list of frame desc mat files
                for frame in sorted(r_map[uid]):
                    ifp, dfp, r = r_map[uid][frame]

                    # Descriptor generation may have failed for this UID
                    try:
                        i_shape, d_shape = r.get()
                    except RuntimeError as ex:
                        self._log.warning('Descriptor generation failed for '
                                          'frame %d in video UID[%s]: %s',
                                          frame, uid, str(ex))
                        r_map[uid] = None
                        continue

                    if d_width is None and d_shape[0] != 0:
                        i_width = i_shape[1]
                        d_width = d_shape[1]

                    # Skip if there were no descriptors generated for this
                    # frame
                    if d_shape[1] == 0:
                        continue

                    video_info_mat_fps.append(ifp)
                    video_desc_mat_fps.append(dfp)
                    video_num_desc += d_shape[0]

                # If combined descriptor height exceeds the per-item limit,
                # generate a random subsample index list
                ssi = None
                if video_num_desc > per_item_limit:
                    ssi = sorted(
                        numpy.random
                             .permutation(video_num_desc)[:per_item_limit]
                    )
                    video_num_desc = len(ssi)

                r_map[uid] = (video_info_mat_fps, video_desc_mat_fps,
                              running_height, ssi)
                running_height += video_num_desc
        pool.join()
        del pool

        with SimpleTimer("Building master descriptor matrices...",
                         self._log.debug):
            master_info = numpy.zeros((running_height, i_width), dtype=float)
            master_desc = numpy.zeros((running_height, d_width), dtype=float)
            tp = multiprocessing.pool.ThreadPool(processes=self.parallel)
            for uid in uids:
                info_fp_list, desc_fp_list, sR, ssi = r_map[uid]
                tp.apply_async(ColorDescriptor_Video._thread_load_matrices,
                               args=(master_info, info_fp_list, sR, ssi))
                tp.apply_async(ColorDescriptor_Video._thread_load_matrices,
                               args=(master_desc, desc_fp_list, sR, ssi))
            tp.close()
            tp.join()

        return master_info, master_desc

    @staticmethod
    def _thread_load_matrices(m, file_list, sR, subsample=None):
        """
        load numpy matrices from files in ``file_list``, concatenating them
        vertically. If a list of row indices is provided in ``subsample`` we
        subsample those rows out of the concatenated matrix. This matrix is then
        inserted into ``m`` starting at row ``sR``.
        """
        c = numpy.load(file_list[0])
        for i in range(1, len(file_list)):
            c = numpy.vstack((c, numpy.load(file_list[i])))
        if subsample:
            c = c[subsample, :]
        m[sR:sR+c.shape[0], :c.shape[1]] = c


# Begin automatic class type creation
valid_descriptor_types = [
    'rgbhistogram',
    'opponenthistogram',
    'huehistogram',
    'nrghistogram',
    'transformedcolorhistogram',
    'colormoments',
    'colormomentinvariants',
    'sift',
    'huesift',
    'hsvsift',
    'opponentsift',
    'rgsift',
    'csift',
    'rgbsift',
]


def _create_image_descriptor_class(descriptor_type_str):
    """
    Create and return a ColorDescriptor class that operates over Image files
    using the given descriptor type.
    """
    assert descriptor_type_str in valid_descriptor_types, \
        "Given ColorDescriptor type was not valid! Given: %s. Expected one " \
        "of: %s" % (descriptor_type_str, valid_descriptor_types)

    # noinspection PyPep8Naming
    class _cd_image_impl (ColorDescriptor_Image):
        def descriptor_type(self):
            """
            :rtype: str
            """
            return descriptor_type_str

    _cd_image_impl.__name__ = "ColorDescriptor_Image_%s" % descriptor_type_str
    return _cd_image_impl


def _create_video_descriptor_class(descriptor_type_str):
    """
    Create and return a ColorDescriptor class that operates over Video files
    using the given descriptor type.
    """
    assert descriptor_type_str in valid_descriptor_types, \
        "Given ColorDescriptor type was not valid! Given: %s. Expected one " \
        "of: %s" % (descriptor_type_str, valid_descriptor_types)

    # noinspection PyPep8Naming
    class _cd_video_impl (ColorDescriptor_Video):
        def descriptor_type(self):
            """
            :rtype: str
            """
            return descriptor_type_str

    _cd_video_impl.__name__ = "ColorDescriptor_Video_%s" % descriptor_type_str
    return _cd_video_impl


# In order to allow multiprocessing, class types must be concretely assigned to
# variables in the module. Dynamic generation causes issues with pickling (the
# default data transmission protocol).

ColorDescriptor_Image_rgbhistogram = \
    _create_image_descriptor_class('rgbhistogram')
ColorDescriptor_Image_opponenthistogram = \
    _create_image_descriptor_class('opponenthistogram')
ColorDescriptor_Image_huehistogram = \
    _create_image_descriptor_class('huehistogram')
ColorDescriptor_Image_nrghistogram = \
    _create_image_descriptor_class('nrghistogram')
ColorDescriptor_Image_transformedcolorhistogram = \
    _create_image_descriptor_class('transformedcolorhistogram')
ColorDescriptor_Image_colormoments = \
    _create_image_descriptor_class('colormoments')
ColorDescriptor_Image_colormomentinvariants = \
    _create_image_descriptor_class('colormomentinvariants')
ColorDescriptor_Image_sift = _create_image_descriptor_class('sift')
ColorDescriptor_Image_huesift = _create_image_descriptor_class('huesift')
ColorDescriptor_Image_hsvsift = _create_image_descriptor_class('hsvsift')
ColorDescriptor_Image_opponentsift = \
    _create_image_descriptor_class('opponentsift')
ColorDescriptor_Image_rgsift = _create_image_descriptor_class('rgsift')
ColorDescriptor_Image_csift = _create_image_descriptor_class('csift')
ColorDescriptor_Image_rgbsift = _create_image_descriptor_class('rgbsift')

ColorDescriptor_Video_rgbhistogram = \
    _create_video_descriptor_class('rgbhistogram')
ColorDescriptor_Video_opponenthistogram = \
    _create_video_descriptor_class('opponenthistogram')
ColorDescriptor_Video_huehistogram = \
    _create_video_descriptor_class('huehistogram')
ColorDescriptor_Video_nrghistogram = \
    _create_video_descriptor_class('nrghistogram')
ColorDescriptor_Video_transformedcolorhistogram = \
    _create_video_descriptor_class('transformedcolorhistogram')
ColorDescriptor_Video_colormoments = \
    _create_video_descriptor_class('colormoments')
ColorDescriptor_Video_colormomentinvariants = \
    _create_video_descriptor_class('colormomentinvariants')
ColorDescriptor_Video_sift = _create_video_descriptor_class('sift')
ColorDescriptor_Video_huesift = _create_video_descriptor_class('huesift')
ColorDescriptor_Video_hsvsift = _create_video_descriptor_class('hsvsift')
ColorDescriptor_Video_opponentsift = \
    _create_video_descriptor_class('opponentsift')
ColorDescriptor_Video_rgsift = _create_video_descriptor_class('rgsift')
ColorDescriptor_Video_csift = _create_video_descriptor_class('csift')
ColorDescriptor_Video_rgbsift = _create_video_descriptor_class('rgbsift')


cd_type_list = [
    ColorDescriptor_Image_rgbhistogram,
    ColorDescriptor_Video_rgbhistogram,
    ColorDescriptor_Image_opponenthistogram,
    ColorDescriptor_Video_opponenthistogram,
    ColorDescriptor_Image_huehistogram,
    ColorDescriptor_Video_huehistogram,
    ColorDescriptor_Image_nrghistogram,
    ColorDescriptor_Video_nrghistogram,
    ColorDescriptor_Image_transformedcolorhistogram,
    ColorDescriptor_Video_transformedcolorhistogram,
    ColorDescriptor_Image_colormoments,
    ColorDescriptor_Video_colormoments,
    ColorDescriptor_Image_colormomentinvariants,
    ColorDescriptor_Video_colormomentinvariants,
    ColorDescriptor_Image_sift,
    ColorDescriptor_Video_sift,
    ColorDescriptor_Image_huesift,
    ColorDescriptor_Video_huesift,
    ColorDescriptor_Image_hsvsift,
    ColorDescriptor_Video_hsvsift,
    ColorDescriptor_Image_opponentsift,
    ColorDescriptor_Video_opponentsift,
    ColorDescriptor_Image_rgsift,
    ColorDescriptor_Video_rgsift,
    ColorDescriptor_Image_csift,
    ColorDescriptor_Video_csift,
    ColorDescriptor_Image_rgbsift,
    ColorDescriptor_Video_rgbsift,
]



from threading import RLock

from smqtk.exceptions import NoClassificationError
from smqtk.representation.classification_element import ClassificationElement


__author__ = "paul.tunison@kitware.com"


class MemoryClassificationElement (ClassificationElement):
    """
    In-memory representation of classification results. This is represented with
    a python dictionary.
    """

    @classmethod
    def is_usable(cls):
        # No external dependencies
        return True

    def __init__(self, type_name, uuid):
        """
        Initialize a new in-memory classification element.

        :param type_name: Name of the type of classifier this classification was
            generated by.
        :type type_name: str

        :param uuid: Unique ID reference of the classification
        :type uuid: collections.Hashable

        """
        super(MemoryClassificationElement, self).__init__(type_name, uuid)

        # dictionary of classification labels and values
        #: :type: None | dict[collections.Hashable, float]
        self._c = None
        # Cannot be pickled. New lock initialized upon pickle/unpickle
        self._c_lock = RLock()

    def __getstate__(self):
        state = {
            'type': self.type_name,
            'uuid': self.uuid
        }
        with self._c_lock:
            state['c'] = self._c
        return state

    def __setstate__(self, state):
        self.type_name = state['type']
        self.uuid = state['uuid']
        self._c_lock = RLock()
        with self._c_lock:
            #: :type: None | dict[collections.Hashable, float]
            self._c = state['c']

    def get_config(self):
        """
        Return a JSON-compliant dictionary that could be passed to this class's
        ``from_config`` method to produce an instance with identical
        configuration.

        :return: JSON type compliant configuration dictionary.
        :rtype: dict

        """
        return {}

    def has_classifications(self):
        """
        :return: If this element has classification information set.
        :rtype: bool
        """
        with self._c_lock:
            return bool(self._c)

    def get_classification(self):
        """
        Get classification result map, returning a label-to-confidence dict.

        We do no place any guarantees on label value types as they may be
        represented in various forms (integers, strings, etc.).

        Confidence values are in the [0,1] range.

        :raises NoClassificationError: No classification labels/confidences yet
            set.

        :return: Label-to-confidence dictionary.
        :rtype: dict[collections.Hashable, float]

        """
        with self._c_lock:
            if self._c:
                return self._c
            else:
                raise NoClassificationError("No classification labels/values")

    def set_classification(self, m=None, **kwds):
        """
        Set the whole classification map for this element. This will strictly
        overwrite the entire label-confidence mapping (vs. updating it)

        Label/confidence values may either be provided via keyword arguments or
        by providing a dictionary mapping labels to confidence values.

        :param m: New labels-to-confidence mapping to set.
        :type m: dict[collections.Hashable, float]

        :raises ValueError: The given label-confidence map was empty.

        """
        m = super(MemoryClassificationElement, self)\
            .set_classification(m, **kwds)
        with self._c_lock:
            self._c = m



from __future__ import print_function

import unittest

import numpy
from sklearn.cluster import MiniBatchKMeans

from smqtk.compute_functions import mb_kmeans_build_apply
from smqtk.representation.descriptor_index.memory import \
    MemoryDescriptorIndex
from smqtk.representation.descriptor_element.local_elements import \
    DescriptorMemoryElement


class TestMBKMClustering (unittest.TestCase):

    def test_clustering_equal_descriptors(self):
        # Test that clusters of descriptor of size  n-features are correctly
        # clustered together.
        print("Creating dummy descriptors")
        n_features = 8
        n_descriptors = 20

        index = MemoryDescriptorIndex()
        c = 0
        for i in range(n_features):
            v = numpy.ndarray((8,))
            v[...] = 0
            v[i] = 1
            for j in range(n_descriptors):
                d = DescriptorMemoryElement('test', c)
                d.set_vector(v)
                index.add_descriptor(d)
                c += 1

        print("Creating test MBKM")
        mbkm = MiniBatchKMeans(n_features, batch_size=12, verbose=True,
                               compute_labels=False, random_state=0)

        # Initial fit with half of index
        d_classes = mb_kmeans_build_apply(index, mbkm, n_descriptors)

        # There should be 20 descriptors per class
        for c in d_classes:
            self.assertEqual(
                len(d_classes[c]),
                n_descriptors,
                "Cluster %s did not have expected number of descriptors "
                "(%d != %d)"
                % (c, n_descriptors, len(d_classes[c]))
            )

            # Each descriptor in each cluster should be equal to the other
            # descriptors in that cluster
            uuids = list(d_classes[c])
            v = index[uuids[0]].vector()
            for uuid in uuids[1:]:
                v2 = index[uuid].vector()
                numpy.testing.assert_array_equal(v, v2,
                                                 "vector in cluster %d did not "
                                                 "match other vectors "
                                                 "(%s != %s)"
                                                 % (c, v, v2))

    def test_empty_index(self):
        # what happens when function given an empty descriptor index
        index = MemoryDescriptorIndex()
        mbkm = MiniBatchKMeans()
        d = mb_kmeans_build_apply(index, mbkm, 0)
        self.assertFalse(d)



import unittest

import mock

from smqtk.algorithms.classifier import ClassifierCollection
from smqtk.exceptions import MissingLabelError
from smqtk.representation.classification_element.memory \
    import MemoryClassificationElement
from smqtk.representation.descriptor_element.local_elements \
    import DescriptorMemoryElement
from smqtk.tests.algorithms.classifier.test_ClassifierAbstract \
    import DummyClassifier


class TestClassifierCollection (unittest.TestCase):

    ##########################################################################
    # Constructor Tests

    def test_new_empty(self):
        ccol = ClassifierCollection()
        self.assertEqual(ccol._label_to_classifier, {})

    def test_new_not_classifier_positional(self):
        # First invalid key should be in error message.
        self.assertRaisesRegexp(
            ValueError,
            "for key 'some label'",
            ClassifierCollection,
            classifiers={'some label': 0}
        )

    def test_new_not_classifier_kwarg(self):
        # First invalid key should be in error message.
        self.assertRaisesRegexp(
            ValueError,
            "for key 'some_label'",
            ClassifierCollection,
            some_label=0
        )

    def test_new_positional(self):
        c = DummyClassifier()
        ccol = ClassifierCollection(classifiers={'a label': c})
        self.assertEqual(ccol._label_to_classifier, {'a label': c})

    def test_new_kwargs(self):
        c = DummyClassifier()
        ccol = ClassifierCollection(a_label=c)
        self.assertEqual(ccol._label_to_classifier, {'a_label': c})

    def test_new_both_pos_and_kwds(self):
        c1 = DummyClassifier()
        c2 = DummyClassifier()
        ccol = ClassifierCollection({'a': c1}, b=c2)
        self.assertEqual(ccol._label_to_classifier,
                         {'a': c1, 'b': c2})

    def test_new_duplicate_label(self):
        c1 = DummyClassifier()
        c2 = DummyClassifier()
        self.assertRaisesRegexp(
            ValueError,
            "Duplicate classifier label 'c'",
            ClassifierCollection,
            {'c': c1},
            c=c2
        )

    ##########################################################################
    # Configuration Tests

    def test_get_default_config(self):
        # Returns a non-empty dictionary with just the example key. Contains
        # a sub-dictionary that would container the implementation
        # specifications.
        c = ClassifierCollection.get_default_config()

        # Should just contain the default example
        self.assertEqual(len(c), 1)
        self.assertIn('__example_label__', c.keys())
        # Should be a plugin config after this.
        self.assertIn('type', c['__example_label__'])

    def test_get_config_empty(self):
        # The config coming out of an empty collection should be an empty
        # dictionary.
        ccol = ClassifierCollection()
        self.assertEqual(ccol.get_config(), {})

    def test_get_config_with_stuff(self):
        c1 = DummyClassifier()
        c2 = DummyClassifier()
        ccol = ClassifierCollection({'a': c1}, b=c2)
        # dummy returns {} config.
        self.assertEqual(
            ccol.get_config(),
            {
                'a': {'DummyClassifier': {}, 'type': 'DummyClassifier'},
                'b': {'DummyClassifier': {}, 'type': 'DummyClassifier'},
            }
        )

    def test_from_config_empty(self):
        ccol = ClassifierCollection.from_config({})
        self.assertEqual(ccol._label_to_classifier, {})

    def test_from_config_skip_example_key(self):
        # If the default example is left in the config, it should be skipped.
        # The string chosen for the example key should be unlikely to be used
        # in reality.
        ccol = ClassifierCollection.from_config({
            '__example_label__':
                'this should be skipped regardless of content'
        })
        self.assertEqual(ccol._label_to_classifier, {})

    @mock.patch('smqtk.algorithms.classifier._classifier_collection'
                '.get_classifier_impls')
    def test_from_config_with_content(self, m_get_impls):
        # Mocking implementation getter to only return the dummy
        # implementation.
        m_get_impls.side_effect = lambda: {
            'DummyClassifier': DummyClassifier,
        }

        ccol = ClassifierCollection.from_config({
            'a': {'DummyClassifier': {}, 'type': 'DummyClassifier'},
            'b': {'DummyClassifier': {}, 'type': 'DummyClassifier'},
        })
        self.assertEqual(
            # Using sort because return from ``keys()`` has no guarantee on
            # order.
            sorted(ccol._label_to_classifier.keys()), ['a', 'b']
        )
        self.assertIsInstance(ccol._label_to_classifier['a'], DummyClassifier)
        self.assertIsInstance(ccol._label_to_classifier['b'], DummyClassifier)

    ##########################################################################
    # Accessor Method Tests

    def test_size_len(self):
        ccol = ClassifierCollection()
        self.assertEqual(ccol.size(), 0)
        self.assertEqual(len(ccol), 0)

        ccol = ClassifierCollection(
            a=DummyClassifier(),
            b=DummyClassifier(),
        )
        self.assertEqual(ccol.size(), 2)
        self.assertEqual(len(ccol), 2)

    def test_labels_empty(self):
        ccol = ClassifierCollection()
        self.assertEqual(ccol.labels(), set())

    def test_labels(self):
        ccol = ClassifierCollection(
            classifiers={
                'b': DummyClassifier(),
            },
            a=DummyClassifier(),
            label2=DummyClassifier(),
        )
        self.assertEqual(ccol.labels(), {'a', 'b', 'label2'})

    def test_add_classifier_not_classifier(self):
        # Attempt adding a non-classifier instance
        ccol = ClassifierCollection()
        # The string 'b' is not a classifier instance.
        self.assertRaisesRegexp(
            ValueError,
            "Not given a Classifier instance",
            ccol.add_classifier,
            'a', 'b'
        )

    def test_add_classifier_duplicate_label(self):
        ccol = ClassifierCollection(a=DummyClassifier())
        self.assertRaisesRegexp(
            ValueError,
            "Duplicate label provided: 'a'",
            ccol.add_classifier,
            'a', DummyClassifier()
        )

    def test_add_classifier(self):
        ccol = ClassifierCollection()
        self.assertEqual(ccol.size(), 0)

        c = DummyClassifier()
        ccol.add_classifier('label', c)
        self.assertEqual(ccol.size(), 1)
        self.assertEqual(ccol._label_to_classifier['label'], c)

    def test_get_classifier_bad_label(self):
        c = DummyClassifier()
        ccol = ClassifierCollection(a=c)
        self.assertRaises(
            KeyError,
            ccol.get_classifier,
            'b'
        )

    def test_get_classifier(self):
        c = DummyClassifier()
        ccol = ClassifierCollection(a=c)
        self.assertEqual(ccol.get_classifier('a'), c)

    def test_remove_classifier_bad_label(self):
        c = DummyClassifier()
        ccol = ClassifierCollection(a=c)
        self.assertRaises(
            KeyError,
            ccol.remove_classifier, 'b'
        )

    def test_remove_classifier(self):
        c = DummyClassifier()
        ccol = ClassifierCollection(a=c)
        ccol.remove_classifier('a')
        self.assertEqual(ccol._label_to_classifier, {})

    ##########################################################################
    # Classification Method Tests

    def test_classify(self):
        ccol = ClassifierCollection({
            'subjectA': DummyClassifier(),
            'subjectB': DummyClassifier(),
        })

        d_v = [0, 1, 2, 3, 4]
        d = DescriptorMemoryElement('memory', '0')
        d.set_vector(d_v)
        result = ccol.classify(d)

        # Should contain one entry for each configured classifier.
        self.assertEqual(len(result), 2)
        self.assertIn('subjectA', result)
        self.assertIn('subjectB', result)
        # Each key should map to a classification element (memory in this case
        # because we're using the default factory)
        self.assertIsInstance(result['subjectA'], MemoryClassificationElement)
        self.assertIsInstance(result['subjectB'], MemoryClassificationElement)
        # We know the dummy classifier outputs "classifications" in a
        # deterministic way: class label is descriptor UUID and classification
        # value is its vector as a list.
        self.assertDictEqual(result['subjectA'].get_classification(),
                             {'0': d_v})
        self.assertDictEqual(result['subjectB'].get_classification(),
                             {'0': d_v})

    def test_classify_subset(self):
        ccol = ClassifierCollection({
            'subjectA': DummyClassifier(),
            'subjectB': DummyClassifier(),
        })

        classifierB = ccol._label_to_classifier['subjectB']
        classifierB.classify = mock.Mock()

        d_v = [0, 1, 2, 3, 4]
        d = DescriptorMemoryElement('memory', '0')
        d.set_vector(d_v)
        result = ccol.classify(d, labels=['subjectA'])

        # Should contain one entry for each requested classifier.
        self.assertEqual(len(result), 1)
        self.assertIn('subjectA', result)
        self.assertNotIn('subjectB', result)
        self.assertFalse(classifierB.classify.called)
        # Each key should map to a classification element (memory in this case
        # because we're using the default factory)
        self.assertIsInstance(result['subjectA'], MemoryClassificationElement)
        # We know the dummy classifier outputs "classifications" in a
        # deterministic way: class label is descriptor UUID and classification
        # value is its vector as a list.
        self.assertDictEqual(result['subjectA'].get_classification(),
                             {'0': d_v})

    def test_classify_empty_subset(self):
        ccol = ClassifierCollection({
            'subjectA': DummyClassifier(),
            'subjectB': DummyClassifier(),
        })

        classifierA = ccol._label_to_classifier['subjectA']
        classifierA.classify = mock.Mock()
        classifierB = ccol._label_to_classifier['subjectB']
        classifierB.classify = mock.Mock()

        d_v = [0, 1, 2, 3, 4]
        d = DescriptorMemoryElement('memory', '0')
        d.set_vector(d_v)
        result = ccol.classify(d, labels=[])

        # Should contain no entries.
        self.assertEqual(len(result), 0)
        self.assertNotIn('subjectA', result)
        self.assertFalse(classifierA.classify.called)
        self.assertNotIn('subjectB', result)
        self.assertFalse(classifierB.classify.called)

    def test_classify_missing_label(self):
        ccol = ClassifierCollection({
            'subjectA': DummyClassifier(),
            'subjectB': DummyClassifier(),
        })

        d_v = [0, 1, 2, 3, 4]
        d = DescriptorMemoryElement('memory', '0')
        d.set_vector(d_v)

        # Should throw a MissingLabelError
        with self.assertRaises(MissingLabelError) as cm:
            ccol.classify(d, labels=['subjectC'])
        self.assertSetEqual(cm.exception.labels, {'subjectC'})

        # Should throw a MissingLabelError
        with self.assertRaises(MissingLabelError) as cm:
            ccol.classify(d, labels=['subjectA', 'subjectC'])
        self.assertSetEqual(cm.exception.labels, {'subjectC'})

        # Should throw a MissingLabelError
        with self.assertRaises(MissingLabelError) as cm:
            ccol.classify(d, labels=['subjectC', 'subjectD'])
        self.assertSetEqual(cm.exception.labels, {'subjectC', 'subjectD'})

        # Should throw a MissingLabelError
        with self.assertRaises(MissingLabelError) as cm:
            ccol.classify(d, labels=['subjectA', 'subjectC', 'subjectD'])
        self.assertSetEqual(cm.exception.labels, {'subjectC', 'subjectD'})



from __future__ import division, print_function
import unittest

import numpy
from six import BytesIO

from smqtk.algorithms.nn_index.hash_index.linear import LinearHashIndex
from smqtk.representation.data_element.memory_element import DataMemoryElement


class TestLinearHashIndex (unittest.TestCase):

    def test_is_usable(self):
        # Should always be true since this impl does no have special deps.
        self.assertTrue(LinearHashIndex.is_usable())

    def test_default_config(self):
        c = LinearHashIndex.get_default_config()
        self.assertEqual(len(c), 1)
        self.assertIsNone(c['cache_element']['type'])

    def test_from_config_no_cache(self):
        # Default config is valid and specifies no cache.
        c = LinearHashIndex.get_default_config()
        i = LinearHashIndex.from_config(c)
        self.assertIsNone(i.cache_element)
        self.assertEqual(i.index, set())

    def test_from_config_with_cache(self):
        c = LinearHashIndex.get_default_config()
        c['cache_element']['type'] = "DataMemoryElement"
        i = LinearHashIndex.from_config(c)
        self.assertIsInstance(i.cache_element, DataMemoryElement)
        self.assertEqual(i.index, set())

    def test_get_config(self):
        i = LinearHashIndex()

        # Without cache element
        expected_c = LinearHashIndex.get_default_config()
        self.assertEqual(i.get_config(), expected_c)

        # With cache element
        i.cache_element = DataMemoryElement()
        expected_c['cache_element']['type'] = 'DataMemoryElement'
        self.assertEqual(i.get_config(), expected_c)

    def test_build_index_no_cache(self):
        i = LinearHashIndex()
        # noinspection PyTypeChecker
        i.build_index([[0, 1, 0],
                       [1, 0, 0],
                       [0, 1, 1],
                       [0, 0, 1]])
        self.assertEqual(i.index, {1, 2, 3, 4})
        self.assertIsNone(i.cache_element)

    def test_build_index_with_cache(self):
        cache_element = DataMemoryElement()
        i = LinearHashIndex(cache_element)
        # noinspection PyTypeChecker
        i.build_index([[0, 1, 0],
                       [1, 0, 0],
                       [0, 1, 1],
                       [0, 0, 1]])
        self.assertEqual(i.index, {1, 2, 3, 4})
        self.assertFalse(cache_element.is_empty())

    def test_build_index_no_input(self):
        i = LinearHashIndex()
        self.assertRaises(
            ValueError,
            i.build_index, []
        )

    def test_update_index_no_input(self):
        i = LinearHashIndex()
        self.assertRaises(
            ValueError,
            i.update_index, []
        )

    def test_update_index_no_index(self):
        # Test calling update index with no existing index.  Should result the
        # same as calling build_index with no index.
        i = LinearHashIndex()
        # noinspection PyTypeChecker
        i.update_index([[0, 1, 0],
                        [1, 0, 0],
                        [0, 1, 1],
                        [0, 0, 1]])
        self.assertEqual(i.index, {1, 2, 3, 4})
        self.assertIsNone(i.cache_element)

    def test_update_index_add_hashes(self):
        i = LinearHashIndex()
        # Build index with some initial hashes
        # noinspection PyTypeChecker
        i.build_index([[0, 0],
                       [0, 1]])
        self.assertSetEqual(i.index, {0, 1})
        # Update index with new stuff
        # noinspection PyTypeChecker
        i.update_index([[1, 0],
                        [1, 1]])
        self.assertSetEqual(i.index, {0, 1, 2, 3})

    def test_remove_from_index_single_not_in_index(self):
        # Test attempting to remove single hash not in the index.
        i = LinearHashIndex()
        i.index = {0, 1, 2}
        self.assertRaises(
            KeyError,
            i.remove_from_index,
            [[1, 0, 0]]  # 4
        )
        self.assertSetEqual(i.index, {0, 1, 2})

    def test_remove_from_index_one_of_many_not_in_index(self):
        # Test attempting to remove hashes where one of them is not in the
        # index.
        i = LinearHashIndex()
        i.index = {0, 1, 2}
        self.assertRaises(
            KeyError,
            i.remove_from_index, [[0, 0],  # 0
                                  [0, 1],  # 1
                                  [1, 1]]  # 3
        )
        # Check that the index has not been modified.
        self.assertSetEqual(i.index, {0, 1, 2})

    def test_remove_from_index(self):
        # Test that actual removal occurs.
        i = LinearHashIndex()
        i.index = {0, 1, 2}
        # noinspection PyTypeChecker
        i.remove_from_index([[0, 0],
                             [1, 0]])
        self.assertSetEqual(i.index, {1})

    def test_nn(self):
        i = LinearHashIndex()
        # noinspection PyTypeChecker
        i.build_index([[0, 1, 0],
                       [1, 1, 0],
                       [0, 1, 1],
                       [0, 0, 1]])
        # noinspection PyTypeChecker
        near_codes, near_dists = i.nn([0, 0, 0], 4)
        self.assertEqual(set(map(tuple, near_codes[:2])),
                         {(0, 1, 0), (0, 0, 1)})
        self.assertEqual(set(map(tuple, near_codes[2:])),
                         {(1, 1, 0), (0, 1, 1)})
        numpy.testing.assert_array_almost_equal(near_dists,
                                                (1/3., 1/3., 2/3., 2/3.))

    def test_save_cache_build_index(self):
        cache_element = DataMemoryElement()
        self.assertTrue(cache_element.is_empty())

        i = LinearHashIndex(cache_element)
        # noinspection PyTypeChecker
        i.build_index([[0, 1, 0],
                       [1, 0, 0],
                       [0, 1, 1],
                       [0, 0, 1]])
        self.assertFalse(cache_element.is_empty())
        # Check byte content
        expected_cache = {1, 2, 3, 4}
        actual_cache = set(numpy.load(BytesIO(cache_element.get_bytes())))
        self.assertSetEqual(expected_cache, actual_cache)

    def test_save_cache_update_index(self):
        cache_element = DataMemoryElement()
        self.assertTrue(cache_element.is_empty())

        i = LinearHashIndex(cache_element)
        # noinspection PyTypeChecker
        i.build_index([[0, 1, 0],   # 2
                       [1, 0, 0]])  # 4
        # noinspection PyTypeChecker
        i.update_index([[0, 1, 1],   # 3
                        [0, 0, 1]])  # 1
        self.assertFalse(cache_element.is_empty())
        # Check byte content
        expected_cache = {1, 2, 3, 4}
        actual_cache = set(numpy.load(BytesIO(cache_element.get_bytes())))
        self.assertSetEqual(expected_cache, actual_cache)

    def test_save_cache_remove_from_index(self):
        # Test that the cache is updated appropriately on a removal.
        cache_element = DataMemoryElement()
        self.assertTrue(cache_element.is_empty())

        i = LinearHashIndex(cache_element)
        # noinspection PyTypeChecker
        i.build_index([[0, 1, 0],   # 2
                       [0, 1, 1],   # 3
                       [1, 0, 0],   # 4
                       [1, 1, 0]])  # 6
        self.assertFalse(cache_element.is_empty())
        self.assertSetEqual(
            set(numpy.load(BytesIO(cache_element.get_bytes()))),
            {2, 3, 4, 6}
        )

        # noinspection PyTypeChecker
        i.remove_from_index([[0, 1, 1],   # 3
                             [1, 0, 0]])  # 4
        self.assertFalse(cache_element.is_empty())
        self.assertSetEqual(
            set(numpy.load(BytesIO(cache_element.get_bytes()))),
            {2, 6}
        )

    def test_save_cache_readonly_build_index(self):
        ro_cache = DataMemoryElement(readonly=True)
        i = LinearHashIndex(ro_cache)
        self.assertRaisesRegexp(
            ValueError,
            "is read-only",
            i.build_index,
            [[0, 1, 0],
             [1, 0, 0],
             [0, 1, 1],
             [0, 0, 1]]
        )

    def test_save_cache_readonly_update_index(self):
        ro_cache = DataMemoryElement(readonly=True)
        i = LinearHashIndex(ro_cache)
        self.assertRaisesRegexp(
            ValueError,
            "is read-only",
            i.update_index,
            [[0, 1, 0],
             [1, 0, 0],
             [0, 1, 1],
             [0, 0, 1]]
        )

    def test_load_cache(self):
        cache_element = DataMemoryElement()
        i1 = LinearHashIndex(cache_element)
        # noinspection PyTypeChecker
        i1.build_index([[0, 1, 0],
                        [1, 0, 0],
                        [0, 1, 1],
                        [0, 0, 1]])

        # load called on initialization.
        i2 = LinearHashIndex(cache_element)

        self.assertEqual(i1.cache_element, i2.cache_element)
        self.assertEqual(i1.index, i2.index)



"""
Collection of higher level functions to perform operational tasks.

Some day, this module could have a companion module containing the CLI logic
for these functions instead of scripts in ``<source>/bin/scripts``.

"""
import collections
import logging

import numpy
import six

from smqtk.utils import (
    bin_utils,
    bit_utils,
    parallel,
)


def compute_many_descriptors(data_elements, descr_generator, descr_factory,
                             descr_index, batch_size=None, overwrite=False,
                             procs=None, **kwds):
    """
    Compute descriptors for each data element, yielding
    (DataElement, DescriptorElement) tuple pairs in the order that they were
    input.

    *Note:* **This function currently only operated over images due to the
    specific data validity check/filter performed.*

    :param data_elements: Iterable of DataElement instances of files to
        work on.
    :type data_elements: collections.Iterable[smqtk.representation.DataElement]

    :param descr_generator: DescriptorGenerator implementation instance
        to use to generate descriptor vectors.
    :type descr_generator: smqtk.algorithms.DescriptorGenerator

    :param descr_factory: DescriptorElement factory to use when producing
        descriptor vectors.
    :type descr_factory: smqtk.representation.DescriptorElementFactory

    :param descr_index: DescriptorIndex instance to add generated descriptors
        to. When given a non-zero batch size, we add descriptors to the given
        index in batches of that size. When a batch size is not given, we add
        all generated descriptors to the index after they have been generated.
    :type descr_index: smqtk.representation.DescriptorIndex

    :param batch_size: Optional number of elements to asynchronously compute
        at a time. This is useful when it is desired for this function to yield
        results before all descriptors have been computed, yet still take
        advantage of any batch asynchronous computation optimizations a
        particular DescriptorGenerator implementation may have. If this is 0 or
        None (false-evaluating), this function blocks until all descriptors have
        been generated.
    :type batch_size: None | int | long

    :param overwrite: If descriptors from a particular generator already exist
        for particular data, re-compute the descriptor for that data and set
        into the generated DescriptorElement.
    :type overwrite: bool

    :param procs: Tell the DescriptorGenerator to use a specific number of
        threads/cores.
    :type procs: None | int

    :param kwds: Remaining keyword-arguments that are to be passed into the
        ``compute_descriptor_async`` function on the descriptor generator.
    :type kwds: dict

    :return: Generator that yields (DataElement, DescriptorElement) for each
        data element given, in the order they were provided.
    :rtype: collections.Iterable[(smqtk.representation.DataElement,
                                  smqtk.representation.DescriptorElement)]

    """
    log = logging.getLogger(__name__)

    # Capture of generated elements in order of generation
    #: :type: deque[smqtk.representation.DataElement]
    de_deque = collections.deque()

    # Counts for logging
    total = 0
    unique = 0

    def iter_capture_elements():
        for d in data_elements:
            de_deque.append(d)
            yield d

    if batch_size:
        log.debug("Computing in batches of size %d", batch_size)

        batch_i = 0

        for _ in iter_capture_elements():
            # elements captured ``de_deque`` in iter_capture_elements

            if len(de_deque) == batch_size:
                batch_i += 1
                log.debug("Computing batch %d", batch_i)

                total += len(de_deque)
                m = descr_generator.compute_descriptor_async(
                    de_deque, descr_factory, overwrite, procs, **kwds
                )
                unique += len(m)
                log.debug("-- Processed %d so far (%d total data elements "
                          "input)", unique, total)

                log.debug("-- adding to index")
                descr_index.add_many_descriptors(six.itervalues(m))

                log.debug("-- yielding generated descriptor elements")
                for e in de_deque:
                    # noinspection PyProtectedMember
                    yield e, m[e.uuid()]

                de_deque.clear()

        if len(de_deque):
            log.debug("Computing final batch of size %d",
                      len(de_deque))

            total += len(de_deque)
            m = descr_generator.compute_descriptor_async(
                de_deque, descr_factory, overwrite, procs, **kwds
            )
            unique += len(m)
            log.debug("-- Processed %d so far (%d total data elements "
                      "input)", unique, total)

            log.debug("-- adding to index")
            descr_index.add_many_descriptors(six.itervalues(m))

            log.debug("-- yielding generated descriptor elements")
            for de in de_deque:
                # noinspection PyProtectedMember
                yield de, m[de.uuid()]

    else:
        log.debug("Using single async call")

        # Just do everything in one call
        log.debug("Computing descriptors")
        m = descr_generator.compute_descriptor_async(
            iter_capture_elements(), descr_factory,
            overwrite, procs, **kwds
        )

        log.debug("Adding to index")
        descr_index.add_many_descriptors(six.itervalues(m))

        log.debug("yielding generated elements")
        for de in de_deque:
            yield de, m[de.uuid()]


def compute_hash_codes(uuids, index, functor, report_interval=1.0, use_mp=False,
                       ordered=False):
    """
    Given an iterable of DescriptorElement UUIDs, asynchronously access them
    from the given ``index``, asynchronously compute hash codes via ``functor``
    and convert to an integer, yielding (UUID, hash-int) pairs.

    :param uuids: Sequence of UUIDs to process
    :type uuids: collections.Iterable[collections.Hashable]

    :param index: Descriptor index to pull from.
    :type index: smqtk.representation.descriptor_index.DescriptorIndex

    :param functor: LSH hash code functor instance
    :type functor: smqtk.algorithms.LshFunctor

    :param report_interval: Frequency in seconds at which we report speed and
        completion progress via logging. Reporting is disabled when logging
        is not in debug and this value is greater than 0.
    :type report_interval: float

    :param use_mp: If multiprocessing should be used for parallel
        computation vs. threading. Reminder: This will copy currently loaded
        objects onto worker processes (e.g. the given index), which could lead
        to dangerously high RAM consumption.
    :type use_mp: bool

    :param ordered: If the element-hash value pairs yielded are in the same
        order as element UUID values input. This function should be slightly
        faster when ordering is not required.
    :type ordered: bool

    :return: Generator instance yielding (DescriptorElement, int) value pairs.

    """
    # TODO: parallel map fetch elements from index?
    #       -> separately from compute

    def get_hash(u):
        v = index.get_descriptor(u).vector()
        return u, bit_utils.bit_vector_to_int_large(functor.get_hash(v))

    # Setup log and reporting function
    log = logging.getLogger(__name__)

    if log.getEffectiveLevel() > logging.DEBUG or report_interval <= 0:
        def log_func(*_, **__):
            return
        log.debug("Not logging progress")
    else:
        log.debug("Logging progress at %f second intervals", report_interval)
        log_func = log.debug

    log.debug("Starting computation")
    reporter = bin_utils.ProgressReporter(log_func, report_interval)
    reporter.start()
    for uuid, hash_int in parallel.parallel_map(get_hash, uuids,
                                                ordered=ordered,
                                                use_multiprocessing=use_mp):
        yield (uuid, hash_int)
        # Progress reporting
        reporter.increment_report()

    # Final report
    reporter.report()


def mb_kmeans_build_apply(index, mbkm, initial_fit_size):
    """
    Build the MiniBatchKMeans centroids based on the descriptors in the given
    index, then predicting descriptor clusters with the final result model.

    If the given index is empty, no fitting or clustering occurs and an empty
    dictionary is returned.

    :param index: Index of descriptors
    :type index: smqtk.representation.DescriptorIndex

    :param mbkm: Scikit-Learn MiniBatchKMeans instead to train and then use for
        prediction
    :type mbkm: sklearn.cluster.MiniBatchKMeans

    :param initial_fit_size: Number of descriptors to run an initial fit with.
        This brings the advantage of choosing a best initialization point from
        multiple.
    :type initial_fit_size: int

    :return: Dictionary of the cluster label (integer) to the set of descriptor
        UUIDs belonging to that cluster.
    :rtype: dict[int, set[collections.Hashable]]

    """
    log = logging.getLogger(__name__)

    ifit_completed = False
    k_deque = collections.deque()
    d_fitted = 0

    log.info("Getting index keys (shuffled)")
    index_keys = sorted(six.iterkeys(index))
    numpy.random.seed(mbkm.random_state)
    numpy.random.shuffle(index_keys)

    def parallel_iter_vectors(descriptors):
        """ Get the vectors for the descriptors given.
        Not caring about order returned.
        """
        return parallel.parallel_map(lambda d: d.vector(), descriptors,
                                     use_multiprocessing=False)

    def get_vectors(k_iter):
        """ Get numpy array of descriptor vectors (2D array returned) """
        return numpy.array(list(
            parallel_iter_vectors(index.get_many_descriptors(k_iter))
        ))

    log.info("Collecting iteratively fitting model")
    rps = [0] * 7
    for i, k in enumerate(index_keys):
        k_deque.append(k)
        bin_utils.report_progress(log.debug, rps, 1.)

        if initial_fit_size and not ifit_completed:
            if len(k_deque) == initial_fit_size:
                log.info("Initial fit using %d descriptors", len(k_deque))
                log.info("- collecting vectors")
                vectors = get_vectors(k_deque)
                log.info("- fitting model")
                mbkm.fit(vectors)
                log.info("- cleaning")
                d_fitted += len(vectors)
                k_deque.clear()
                ifit_completed = True
        elif len(k_deque) == mbkm.batch_size:
            log.info("Partial fit with batch size %d", len(k_deque))
            log.info("- collecting vectors")
            vectors = get_vectors(k_deque)
            log.info("- fitting model")
            mbkm.partial_fit(vectors)
            log.info("- cleaning")
            d_fitted += len(k_deque)
            k_deque.clear()

    # Final fit with any remaining descriptors
    if k_deque:
        log.info("Final partial fit of size %d", len(k_deque))
        log.info('- collecting vectors')
        vectors = get_vectors(k_deque)
        log.info('- fitting model')
        mbkm.partial_fit(vectors)
        log.info('- cleaning')
        d_fitted += len(k_deque)
        k_deque.clear()

    log.info("Computing descriptor classes with final KMeans model")
    mbkm.verbose = False
    d_classes = collections.defaultdict(set)
    d_uv_iter = parallel.parallel_map(lambda d: (d.uuid(), d.vector()),
                                      index,
                                      use_multiprocessing=False,
                                      name="uv-collector")
    # TODO: Batch predict call inputs to something larger than one at a time.
    d_uc_iter = parallel.parallel_map(
        lambda u_v: (u_v[0], mbkm.predict(u_v[1][numpy.newaxis, :])[0]),
        d_uv_iter,
        use_multiprocessing=False,
        name="uc-collector")
    rps = [0] * 7
    for uuid, c in d_uc_iter:
        d_classes[c].add(uuid)
        bin_utils.report_progress(log.debug, rps, 1.)
    rps[1] -= 1
    bin_utils.report_progress(log.debug, rps, 0)

    return d_classes
